!-------------------------------------------------------------------------------
! LiSDAS v2.7
!
! Copyright 2010
!   YOSHIDA, Ryo,
!   SAITO, Masaya,
!   NAGAO, Hiromichi,
!   NAKANO, Shin'ya,
!   HIGUCHI, Tomoyuki,
!   The Institute of Statistical Mathematics
!
! Dual licensed under the MIT or GPL Version 3 licenses.
! See 'LICENSE.MIT' for the definition of the MIT license, 
! and 'COPYING.GPL' for the definition of the GPL license.
!-------------------------------------------------------------------------------
#define assert(expr) if (.not.(expr)) then; write(*,*) 'diff.f90', ':', __LINE__, ': Assertion: ', "expr", ' failed.'; stop; end if
module M_LiSDAS
  implicit none

! --- Constant parameters ---
!/The character code of TAB used in making output texts
  character, parameter :: tab = char(9)
!/Array allocation:
  !@ Number of time points of time course observations and
  !  state variables (endogeneous variables in model)
  integer, parameter :: max_time_Data = 1000
  integer, parameter :: max_time_StateVariable = 1000
  
!/ Array allocation parameter:
  !@ Number of state variables in model
  integer, parameter :: max_num_Variable = 300

!/ Array allocation parameter:
  !@ Number of particles used in parameter search
  integer, parameter :: max_num_Particle = 3000000

!/ Array allocation parameter:
  !@ Numbers of reactions in the pathway model
  ! Number of reactions in "Reaction Elements" attribute
  integer, parameter :: max_num_Reg_Elements = 500
  ! Number of reactions in "Reaction Modules" attribute
  integer, parameter :: max_num_Reg_Modules = 500 
  ! Number of columns of 'Reg_Elements' and 'Reg_Modules'
  integer, parameter ::  max_tmp = 20

!/ Array allocation parameter:
  !@ Number of parameters used in the specification of prior distributions
  ! Number of parameters in "Elements" attribute
  integer, parameter :: max_num_Para1 = 300
  ! Number of parameters in "Modules" attribute
  integer, parameter :: max_num_Para2 = 300

!/ Number of particles saved at 'save_simulation_*.log'
  integer, parameter :: nsave_Particle = 128

!/ Number of Gaussian radius functions for nonparametric regression
  integer, parameter :: max_nbasisGauss = 30

!/ Number of duration times with increasing/decreasing trends
  integer, parameter :: nDecompTime = 20

! /Additional parameters relavant to the DPM 
  integer, parameter :: max_num_Particle_DPM = 4096

!/ State variables
  type T_SV
    integer :: num_Variable, time_StateVariable
    ! Names of state variables
    character(len=30) :: Name_Variable(max_num_Variable)
    ! State variable IDs
    integer :: ID_Variable(max_num_Variable)
    real(8) :: StateVariable(max_num_Variable, max_time_StateVariable)
  end type

!/ State variables (tentative)
  type T_SV2
    integer :: num_Variable, time_StateVariable
    ! Names of state variables
    character(len=30) :: Name_Variable(max_num_Variable)
    ! State variable IDs
    integer :: ID_Variable(max_num_Variable)
    real(8) :: StateVariable(max_num_Particle_DPM,&
                 max_num_Variable, max_time_StateVariable)
  end type

!/ Time course data and state variables
  type T_Dat
      real(8) :: Data(max_num_Variable, max_time_Data)
    ! Set of time points in 'Data', 
    !   which are mapped the simulation time steps
      integer :: ID_Time_Data(max_time_Data)
    ! 0/1 variables indicating missing/non-missing endogeneous variables
      integer :: ID_MissVariable( max_num_Variable )
      integer :: time_Data
  end type

!/ Matrices defining structure and reaction kinetics of model
  type T_Mdl
    integer :: num_Reg_Elements, num_Reg_Modules 

    !@ Michaelis-Menten-like kinetic values and steepness parameters 
    ! of hill curve functions are saved in the 1st and 2nd columns
    real(8) :: Para_Reg_Elements(max_num_Reg_Elements, max_tmp)

    !@ Coefficients and their signs are saved in the 1st and 2nd columns
    real(8) :: Para_Reg_Modules(max_num_Reg_Modules, max_tmp)

    !@ (1)Element IDs, (2)kinetic law IDs, (3)reactant state variables
    integer :: List_Reg_Elements(max_num_Reg_Elements, max_tmp)

    !@ (1)Module IDs, (2)product variables, (3)parameter ID, 
    !  (4)number of reactant variables 
    !  (5-Last) Element IDs associated to the reaction module
    integer :: List_Reg_Modules(max_num_Reg_Modules, max_tmp)

    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    ! An additional table to do simulation effectively.
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    ! tables maps an disjoint element (discussed lator) to an element and a module
    !   map_e2m(1,*): e_disjoint -> element
    !   map_e2m(2,*): e_disjoint -> module
    integer map_e2m(2,max_num_Reg_Elements), nn
    ! Note: the sum of the numbers of elements in respective modules 
    !   is larger than the number of elements defined in Element section,
    !   since an elements can belong to multiple modules. 
    !   So this declaration may cause the out of bound error, 
    !   if max_num_Reg_Elements is chosen to be equal to its actual size.

    ! tables to refer directorly from e_disjoint to respective items
    integer iList(max_num_Reg_Elements, 5)        ! e_disj -> element
    integer iModuleList(max_num_Reg_Elements, 5)  ! e_disj -> module
    integer iReactantList(max_num_Reg_Elements, 5)! e_disj -> state variable
    integer num_List(5)                           ! sizes

    !/ characters involving lists of model-defining parameters
    character strElements(max_num_Reg_Elements, 4) * 30
    character strModules(max_num_Reg_Modules, 3) * 30 !(1) Product, (2) Reaction, (3) Reactants
    character strParaElements(max_num_Para1, 4)  * 30
    character strParaModules(max_num_Para2) * 50
      character suffix * 30     ! Analysis identifier
      character Name_Model * 30    ! Name of model
  end type

  type T_Mdl2
    integer :: num_Reg_Elements, num_Reg_Modules 

    !@ Michaelis-Menten-like kinetic values and steepness parameters 
    ! of hill curve functions are saved in the 1st and 2nd columns
    real(8) :: Para_Reg_Elements(max_num_Particle_DPM, max_num_Reg_Elements, max_tmp)

    !@ Coefficients and their signs are saved in the 1st and 2nd columns
    real(8) :: Para_Reg_Modules(max_num_Particle_DPM, max_num_Reg_Modules, max_tmp)

    !@ (1)Element IDs, (2)kinetic law IDs, (3)reactant state variables
    integer :: List_Reg_Elements(max_num_Reg_Elements, max_tmp)

    !@ (1)Module IDs, (2)product variables, (3)parameter ID, 
    !  (4)number of reactant variables 
    !  (5-Last) Element IDs associated to the reaction module
    integer :: List_Reg_Modules(max_num_Reg_Modules, max_tmp)

    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    ! An additional table to do simulation effectively.
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    ! tables maps an disjoint element (discussed lator) to an element and a module
    !   map_e2m(1,*): e_disjoint -> element
    !   map_e2m(2,*): e_disjoint -> module
    integer map_e2m(2,max_num_Reg_Elements), nn
    ! Note: the sum of the numbers of elements in respective modules 
    !   is larger than the number of elements defined in Element section,
    !   since an elements can belong to multiple modules. 
    !   So this declaration may cause the out of bound error, 
    !   if max_num_Reg_Elements is chosen to be equal to its actual size.

    ! tables to refer directorly from e_disjoint to respective items
    integer iList(max_num_Reg_Elements, 5)        ! e_disj -> element
    integer iModuleList(max_num_Reg_Elements, 5)  ! e_disj -> module
    integer iReactantList(max_num_Reg_Elements, 5)! e_disj -> state variable
    integer num_List(5)                           ! sizes

    !/ characters involving lists of model-defining parameters
    character strElements(max_num_Reg_Elements, 4) * 30
    character strModules(max_num_Reg_Modules, 3) * 30 !(1) Product, (2) Reaction, (3) Reactants
    character strParaElements(max_num_Para1, 4)  * 30
    character strParaModules(max_num_Para2) * 50
      character suffix * 30     ! Analysis identifier
      character Name_Model * 30    ! Name of model
  end type

!    / Prior specification matrices for "Reaction Elements", 
  type T_Pri
      real(8) :: Prior_ParaElements(max_num_Para1, 4)
    ! @ (1) Mean, (2) variance
      real(8) :: Prior_ParaModules(max_num_Para2, 2)

    ! @ (1) Mean, (2) Variance    
      real(8) :: Prior_Inits(max_num_Variable, 2)

    ! @'Prior_ParaElements' is mapped to 'Reg_Elements' 
    !   (IDs saved at the first column)
      integer :: ID_Prior_ParaElements(max_num_Para1)

    ! @'Prior_ParaModules' is mapped to 'Reg_Modules'
    !   (IDs saved at the third column)
      integer :: ID_Prior_ParaModules(max_num_Para2)

    ! @ 'Prior_Inits' is mapped to 'sv%StateVariable'    
      integer :: ID_Prior_Inits(max_num_Variable)

    ! size of Para_Elements and Para_Modules
    integer :: num_Para1, num_Para2
  end type

!/ Maintained particles for the parameters in "Reaction Elements", 
!  "Reaction Modules", and "Initial States"
  type T_Par
    real(8) :: Particle_ParaElements(max_num_Reg_Elements, 2 )
    real(8) :: Particle_ParaModules(max_num_Reg_Modules)
    real(8)    :: Particle_Inits(max_num_Variable)
  end type
  type T_ParVec
    type(T_Par) :: o(nsave_Particle)
    integer :: num_Particle
  end type
contains
  ! Number to string without redandant spaces
  ! integer -> string
  character(16) function sI(i)
    integer :: i
    write(sI,*) i 
    sI = adjustl(sI) 
  end function
  ! real(8) -> string
  character(32) function sR8(x)
    real(8) :: x
    write(sR8,*) x
    sR8 = adjustl(sR8) 
  end function

  subroutine Wrap_Run_BioSimulator&
    (mdl, sv, time_interval,includeNoise, ix, Variance_System)
    type (T_Mdl) :: mdl
    type (T_SV)  :: sv
    real(8)      :: time_interval
    integer      :: includeNoise
    integer      :: ix
    real(8)      :: Variance_System(*)

      call Run_BioSimulator(&
          !-- Reaction functions
        mdl%List_Reg_Elements, mdl%Para_Reg_Elements,&
        mdl%List_Reg_Modules, mdl%Para_Reg_Modules,&
      mdl%num_Reg_Elements, mdl%num_Reg_Modules,&
      max_num_Reg_Elements, max_num_Reg_Modules, max_tmp,&
      !-- Parameters for simulation
      time_interval, sv%time_StateVariable, max_time_StateVariable,&
      includeNoise, ix, Variance_System,&
          !-- Others
      sv%num_Variable, max_num_Variable,&
          !-- Outputs
      sv%StateVariable&
     )
  end subroutine

  subroutine Wrap_Run_BioSimulator_version3&
    (mdl, sv, time_interval,includeNoise, ix, Variance_System)
    type (T_Mdl) :: mdl
    type (T_SV)  :: sv
    real(8)      :: time_interval
    integer      :: includeNoise
    integer      :: ix
    real(8)      :: Variance_System(*)

    call Run_BioSimulator_version3(&
      !-- Reaction functions
      mdl%List_Reg_Elements, mdl%Para_Reg_Elements,&
      mdl%List_Reg_Modules, mdl%Para_Reg_Modules,&
      mdl%num_Reg_Elements, mdl%num_Reg_Modules,&
      max_num_Reg_Elements, max_num_Reg_Modules, max_tmp,&
      !-- Parameters for simulation
      time_interval, sv%time_StateVariable, max_time_StateVariable,&
      includeNoise, ix, Variance_System,&
      !-- Others
      sv%num_Variable, max_num_Variable,&
      !-- Outputs
      sv%StateVariable&
      )
  end subroutine

  subroutine Setup_Run_BioSimulator_version2(mdl)
    type (T_Mdl) :: mdl
    integer :: i, j, ii, ide, nc
    mdl%nn = 0
    do i = 1, mdl%num_Reg_Elements
      ide = mdl%List_Reg_Elements(i,1)
      do j = 1, mdl%num_Reg_Modules
        nc = mdl%List_Reg_Modules(j,4)
        do ii = 1, nc
          if ( mdl%List_Reg_Modules(j,4+ii).eq.ide ) then
            mdl%nn = mdl%nn + 1
            mdl%map_e2m(1,mdl%nn) = i
            mdl%map_e2m(2,mdl%nn) = j
          end if
        end do
      end do
    end do
  end subroutine

  subroutine Run_BioSimulator_version2&
    (mdl, sv, time_interval,includeNoise, ix, Variance_System)
    type (T_Mdl) :: mdl
    type (T_SV)  :: sv
    real(8)      :: time_interval
    integer      :: includeNoise
    integer      :: ix
    real(8)      :: Variance_System(*)

    real(8) :: a(max_num_Reg_Modules)
    integer :: time, i, ie, im, ID_Reg, ID_Reactant, ID_Product
    real(8), external :: rgauss
  
    do time = 2, sv%time_StateVariable
      do i = 1, mdl%num_Reg_Modules
        a(i) = time_interval*mdl%Para_Reg_Modules(i,1)*mdl%Para_Reg_Modules(i,2)
      end do
  
      do i = 1, sv%num_Variable
        sv%StateVariable(i,time) = 0.0d0
      end do
  
      do i = 1, mdl%nn
        ie = mdl%map_e2m(1,i)
        im = mdl%map_e2m(2,i)
        ID_Reg = mdl%List_Reg_Elements(ie,2)
        ID_Reactant = mdl%List_Reg_Elements(ie,3)
         
        if (ID_Reg .EQ. 1) then ! Differentiation over time
          a(im) = a(im) * sv%StateVariable(ID_Reactant,time-1)/time_interval

        else if (ID_Reg .GE. 3 .AND. ID_Reg .LE. 4) then
          a(im) = a(im) * sv%StateVariable(ID_Reactant,time-1)
 
        else if (ID_Reg .EQ. 18) then
          a(im) = a(im) /&
                ( 1.0d0&
                  + (mdl%Para_Reg_Elements(ie,1) / sv%StateVariable(ID_Reactant,time-1))&
                      ** mdl%Para_Reg_Elements(ie,2))
        else if (ID_Reg .EQ. 19) then
          a(im) = a(im) /&
                ( 1.0d0&
                  + (sv%StateVariable(ID_Reactant,time-1) / mdl%Para_Reg_Elements(ie,1))&
                      ** mdl%Para_Reg_Elements(ie,2))   
        end if
      end do
  
      do i = 1, mdl%num_Reg_Modules
        ID_Product = mdl%List_Reg_Modules(i,2)
        sv%StateVariable(ID_Product,time) =&
          sv%StateVariable(ID_Product,time) + a(i)
      end do
  
      if (includeNoise .EQ. 1) then
        do i = 1, sv%num_Variable
          sv%StateVariable(i,time) = sv%StateVariable(i,time) +&
             dsqrt(Variance_System(i)) * rgauss(ix)
        end do
      end if
  
      do i = 1, sv%num_Variable
        if (sv%StateVariable(i,time) .le. 0.0d0)&
          sv%StateVariable(i,time) = sv%StateVariable(i,time-1)
      end do
    end do
  end subroutine

  subroutine Setup_Run_BioSimulator_version2a(mdl)
    type (T_Mdl) :: mdl
    integer :: ie, im, ie_disjoint, ID_Reg, ID_Reactant

    call Setup_Run_BioSimulator_version2(mdl)! to make map_e2m(1,*)
    !!!!!!!!!!!! fj add
    mdl%num_List = 0
    do ie_disjoint = 1, mdl%nn ! Note it is not equal to num_Reg_Elements!!
      ie = mdl%map_e2m(1,ie_disjoint)
      im = mdl%map_e2m(2,ie_disjoint)
      ID_Reg = mdl%List_Reg_Elements(ie,2)
      ID_Reactant = mdl%List_Reg_Elements(ie,3)
!
      if (ID_Reg .EQ. 1) then
        mdl%num_List(1) = mdl%num_List(1) + 1
        mdl%iList(mdl%num_List(1), 1) = ie
        mdl%iModuleList(mdl%num_List(1), 1) = im
        mdl%iReactantList(mdl%num_List(1), 1) = ID_Reactant
!
      else if (ID_Reg .EQ. 2) then
        mdl%num_List(2) = mdl%num_List(2) + 1
        mdl%iList(mdl%num_List(2), 2) = ie
        mdl%iModuleList(mdl%num_List(2), 2) = im
        mdl%iReactantList(mdl%num_List(2), 2) = ID_Reactant
!
      else if (ID_Reg .GE. 3 .AND. ID_Reg .LE. 4) then
        mdl%num_List(3) = mdl%num_List(3) + 1
        mdl%iList(mdl%num_List(3), 3) = ie
        mdl%iModuleList(mdl%num_List(3), 3) = im
        mdl%iReactantList(mdl%num_List(3), 3) = ID_Reactant
!
      else if (ID_Reg .EQ. 18) then
        mdl%num_List(4) = mdl%num_List(4) + 1
        mdl%iList(mdl%num_List(4), 4) = ie
        mdl%iModuleList(mdl%num_List(4), 4) = im
        mdl%iReactantList(mdl%num_List(4), 4) = ID_Reactant
      else if (ID_Reg .EQ. 19) then
!
        mdl%num_List(5) = mdl%num_List(5) + 1
        mdl%iList(mdl%num_List(5), 5) = ie
        mdl%iModuleList(mdl%num_List(5), 5) = im
        mdl%iReactantList(mdl%num_List(5), 5) = ID_Reactant
      endif
    end do
  end subroutine 

  subroutine Run_BioSimulator_version2a&
    (mdl, sv, time_interval,includeNoise, ix, Variance_System)
    type (T_Mdl) :: mdl
    type (T_SV)  :: sv
    real(8)      :: time_interval
    integer      :: includeNoise
    integer      :: ix
    real(8)      :: Variance_System(*)

    real(8) :: a(max_num_Reg_Modules)
    integer :: time, i, j, ie, im, ir, ID_Product
    real(8), external :: rgauss
  
    do time = 2, sv%time_StateVariable
      do i = 1, mdl%num_Reg_Modules
        a(i) = time_interval*mdl%Para_Reg_Modules(i,1)*mdl%Para_Reg_Modules(i,2)
      end do
  
      do i = 1, sv%num_Variable
        sv%StateVariable(i,time) = 0.0d0
      end do

      ! kind = 1: differentiation over time
      do j = 1, mdl%num_List(1)
         ie = mdl%iList(j, 1)
         im = mdl%iModuleList(j, 1)
         ir = mdl%iReactantList(j, 1)
         a(im) = a(im) * sv%StateVariable(ir,time-1)/time_interval
      enddo
    
      ! kind = 3:
      do j = 1, mdl%num_List(3)
         ie = mdl%iList(j, 3)
         im = mdl%iModuleList(j, 3)
         ir = mdl%iReactantList(j, 3)
         a(im) = a(im)*sv%StateVariable(ir,time-1)
      end do

      ! kind = 4:
      do j = 1, mdl%num_List(4)
         ie = mdl%iList(j, 4)
         im = mdl%iModuleList(j, 4)
         ir = mdl%iReactantList(j, 4)
         a(im) = a(im) /&
                ( 1.0d0&
                  + (mdl%Para_Reg_Elements(ie,1) / sv%StateVariable(ir,time-1))&
                      ** mdl%Para_Reg_Elements(ie,2))
      end do

      ! kind = 5
      do j = 1, mdl%num_List(5)
         ie = mdl%iList(j, 5)
         im = mdl%iModuleList(j, 5)
         ir = mdl%iReactantList(j, 5)
         a(im) = a(im) /&
                ( 1.0d0&
                  + (sv%StateVariable(ir,time-1) / mdl%Para_Reg_Elements(ie,1))&
                      ** mdl%Para_Reg_Elements(ie,2))   
      end do

      do i = 1, mdl%num_Reg_Modules
        ID_Product = mdl%List_Reg_Modules(i,2)
        sv%StateVariable(ID_Product,time) =&
          sv%StateVariable(ID_Product,time) + a(i)
      end do
  
      if (includeNoise .EQ. 1) then
        do i = 1, sv%num_Variable
          sv%StateVariable(i,time) = sv%StateVariable(i,time) +&
             dsqrt(Variance_System(i)) * rgauss(ix)
        end do
      end if
  
      do i = 1, sv%num_Variable
        if (sv%StateVariable(i,time) .le. 0.0d0)&
          sv%StateVariable(i,time) = sv%StateVariable(i,time-1)
      end do
    end do
  end subroutine

  subroutine Run_BioSimulator_version2b&
    (mdl, sv, time_interval,includeNoise, ix, Variance_System, nth)
    type (T_Mdl2):: mdl
    type (T_SV2) :: sv
    real(8)      :: time_interval
    integer      :: includeNoise
    integer      :: ix
    real(8)      :: Variance_System(*)

    real(8), save :: a(max_num_Particle_DPM, max_num_Reg_Modules)
    integer :: time, i, j, ie, im, ir, ID_Product
    integer :: ith, nth
    real(8), external :: rgauss

    assert(nth .le. size(mdl%Para_Reg_Modules, 1))
    assert(nth .le. size(mdl%Para_Reg_Modules, 1))
    assert(nth .le. size(sv%StateVariable, 1))

    !$omp parallel private(time, i, j, ie, im, ir, ID_Product)&
    !$omp   shared(sv, mdl, time_interval, a)
    do time = 2, sv%time_StateVariable
      do i = 1, mdl%num_Reg_Modules
        !$omp do
        do ith = 1, nth
          a(ith, i) = time_interval*&
            mdl%Para_Reg_Modules(ith,i,1)*mdl%Para_Reg_Modules(ith,i,2)
        end do
        !$omp end do
      end do
  
      do i = 1, sv%num_Variable
        !$omp do
        do ith = 1, nth
          sv%StateVariable(ith,i,time) = 0.0d0
        end do
        !$omp end do
      end do

      ! kind = 1: differentiation over time
      do j = 1, mdl%num_List(1)
        ie = mdl%iList(j, 1)
        im = mdl%iModuleList(j, 1)
        ir = mdl%iReactantList(j, 1)
        !$omp do
        do ith = 1, nth
          a(ith,im) = a(ith,im) * sv%StateVariable(ith,ir,time-1)/time_interval
        end do
        !$omp end do
      enddo
    
      ! kind = 3:
      do j = 1, mdl%num_List(3)
        ie = mdl%iList(j, 3)
        im = mdl%iModuleList(j, 3)
        ir = mdl%iReactantList(j, 3)
        !$omp do
        do ith = 1, nth
          a(ith,im) = a(ith,im)*sv%StateVariable(ith,ir,time-1)
        end do
        !$omp end do
      end do

      ! kind = 4:
      do j = 1, mdl%num_List(4)
        ie = mdl%iList(j, 4)
        im = mdl%iModuleList(j, 4)
        ir = mdl%iReactantList(j, 4)
        !$omp do
        do ith = 1, nth
        a(ith,im) = a(ith,im) /&
                ( 1.0d0&
                  + (mdl%Para_Reg_Elements(ith,ie,1) / sv%StateVariable(ith,ir,time-1))&
                      ** mdl%Para_Reg_Elements(ith,ie,2))
        end do
        !$omp end do
      end do

      ! kind = 5
      do j = 1, mdl%num_List(5)
        ie = mdl%iList(j, 5)
        im = mdl%iModuleList(j, 5)
        ir = mdl%iReactantList(j, 5)
        !$omp do
        do ith = 1, nth
        a(ith,im) = a(ith,im) /&
                ( 1.0d0&
                  + (sv%StateVariable(ith,ir,time-1) / mdl%Para_Reg_Elements(ith,ie,1))&
                      ** mdl%Para_Reg_Elements(ith,ie,2))   
        end do
        !$omp end do
      end do

      do i = 1, mdl%num_Reg_Modules
        ID_Product = mdl%List_Reg_Modules(i,2)
        !$omp do
        do ith = 1, nth
          sv%StateVariable(ith,ID_Product,time) =&
            sv%StateVariable(ith,ID_Product,time) + a(ith,i)
        end do
        !$omp end do
      end do
  
      if (includeNoise .EQ. 1) then
        do i = 1, sv%num_Variable
          !$omp do
          do ith = 1, nth
          sv%StateVariable(ith,i,time) = sv%StateVariable(ith,i,time) +&
             dsqrt(Variance_System(i)) * rgauss(ix)
          end do
          !$omp end do
        end do
      end if
  
      do i = 1, sv%num_Variable
        !$omp do
        do ith = 1, nth
        if (sv%StateVariable(ith,i,time) .le. 0.0d0)&
          sv%StateVariable(ith,i,time) = sv%StateVariable(ith,i,time-1)
        end do
        !$omp end do
      end do
    end do
    !$omp end parallel
  end subroutine



  double precision function Wrap_Gauss_Likelihood(dat, sv, Variance_Measure)&
    result (Gauss_Likelihood)
    type (T_Dat) :: dat
    type (T_SV)  :: sv
    real(8) :: Variance_Measure(max_num_Variable)
    integer :: i, j
    real(8) :: sum, time

    Gauss_Likelihood = 0.0d0
    do i = 1, sv%num_Variable
      sum = 0.0d0
      if (dat%ID_MissVariable(i) .EQ. 1) then
        do j = 1, dat%time_Data
        time = dat%ID_Time_Data(j)
        sum = sum + (dat%Data(i,j) - sv%StateVariable(i,time))**2  
        enddo
        Gauss_Likelihood = Gauss_Likelihood + sum/(2.0d0*Variance_Measure(i))
      end if
      if( Gauss_Likelihood.GE.10d8 ) then
        write(*,*) '---Warning() -GaussLikelihood -overflow'
        stop
      end if
    end do
  end function

  subroutine Wrap_Output_BioSimulator(sv, Output_File)
    type (T_SV) :: sv
    character(len=*) :: Output_File
      call Output_BioSimulator(&
                sv%StateVariable, sv%num_Variable, sv%time_StateVariable,&
              max_time_StateVariable, max_num_Variable,&
              sv%Name_Variable,&
              Output_File&
            )
  end subroutine

  subroutine Wrap_Get_Reactions(mdl, sv, time_interval, pri)
    type (T_Mdl) :: mdl
    type (T_SV)  :: sv
    real(8)      :: time_interval
    type (T_Pri) :: pri

    call Get_Reactions (&
!        -- Reaction functions 
             mdl%List_Reg_Elements, mdl%Para_Reg_Elements,&
             mdl%List_Reg_Modules, mdl%Para_Reg_Modules,&
             mdl%num_Reg_Elements, mdl%num_Reg_Modules,&
             max_num_Reg_Elements, max_num_Reg_Modules, max_tmp,&
!        -- Variables
             sv%ID_Variable, sv%Name_Variable, sv%num_Variable,&
             max_num_Variable,&
!        -- Parameters for simulation
             time_interval, sv%time_StateVariable,&
!        -- Prior distributions    
             pri%Prior_ParaElements, pri%ID_Prior_ParaElements,&
             pri%Prior_ParaModules, pri%ID_Prior_ParaModules,&
             pri%Prior_Inits, pri%ID_Prior_Inits,&
             pri%num_Para1, pri%num_Para2, max_num_Para1, max_num_Para2,&
!        -- Others&
             mdl%suffix, mdl%Name_Model,&
             mdl%strElements, mdl%strModules, mdl%strParaElements, mdl%strParaModules&
         )
  end subroutine

  subroutine Wrap_Output_Reactions&
    (mdl, sv, time_interval, pri&
    ,Variance_System, Variance_Measure, Output_File)
    type (T_Mdl) :: mdl
    type (T_SV)  :: sv
    real(8)      :: time_interval
    type (T_Pri) :: pri
    real(8) :: Variance_Measure(max_num_Variable)
    real(8) :: Variance_System(max_num_Variable)
    character(*) :: Output_File
    call Output_Reactions (&
! -- Reaction functions &
             mdl%List_Reg_Elements, mdl%Para_Reg_Elements,&
             mdl%List_Reg_Modules, mdl%Para_Reg_Modules,&
             mdl%num_Reg_Elements, mdl%num_Reg_Modules, &
             max_num_Reg_Elements, max_num_Reg_Modules, max_tmp, &
!        -- Variables&
             sv%ID_Variable, sv%Name_Variable, sv%num_Variable, &
             max_num_Variable,&
!        -- Parameters for simulation&
             time_interval, sv%time_StateVariable, &
!        -- Prior distributions        &
             pri%Prior_ParaElements, pri%ID_Prior_ParaElements, &
             pri%Prior_ParaModules, pri%ID_Prior_ParaModules,&
             pri%Prior_Inits, pri%ID_Prior_Inits, &
             pri%num_Para1, pri%num_Para2, max_num_Para1, max_num_Para2,&
!        -- Others&
             mdl%suffix, mdl%Name_Model, Output_File,&
             sv%StateVariable, max_time_StateVariable,&
!        -- Variances&
             Variance_System, Variance_Measure,&
             mdl%strElements, mdl%strModules, mdl%strParaElements, mdl%strParaModules&
         )
  end subroutine

  subroutine Sampling_PriorG0(pri, priG0, sv, ix, method)
    type(T_Pri) :: pri, priG0
    type(T_SV)  :: sv
    integer     :: ix
    integer     :: method
    integer, parameter :: itrmax = 10000
    integer :: i, j
    real(8) :: rand1, rand2, rg
    real(8), external :: rgauss

!     Sampling from priors for ParaElements
    do i = 1, priG0%num_Para1
!     Switching levels
      do j = 1, itrmax
        rg = rgauss(ix)
        rand1 = priG0%Prior_ParaElements(i,1)&
              + dsqrt( priG0%Prior_ParaElements(i,3) ) ! rg ! s
        if (rand1 .GE. 0.0d0) exit
      end do
!     Switching levels
      do j = 1, itrmax
        rg = rgauss(ix)
        rand2 = priG0%Prior_ParaElements(i,2)&
              + dsqrt( priG0%Prior_ParaElements(i,4) ) ! rg ! s
        if (rand2 .GE. 0.0d0) exit
      end do
!        
      pri%Prior_ParaElements(i,1) = rand1
      pri%Prior_ParaElements(i,2) = rand2
    end do

!     Sampling from priors for ParaModules
    do i = 1, priG0%num_Para2
      do j = 1, itrmax
        rg = rgauss(ix)
        rand1 = priG0%Prior_ParaModules(i,1)& 
              + dsqrt(priG0%Prior_ParaModules(i,2)) ! rg ! scale_variace
        if (rand1 .GE. 0.0d0) exit
      end do
      pri%Prior_ParaModules(i,1) = rand1
    end do    

!     Sampling from priors for initial state conditions
    do i = 1, sv%num_Variable
      do j = 1, itrmax
        rg = rgauss(ix)
        rand1 = priG0%Prior_Inits(i,1)&
              + dsqrt( priG0%Prior_Inits(i,2) ) ! rg ! scale_variace
        if (rand1 .GE. 0.0d0) exit
      end do
      pri%Prior_Inits(i,1) = rand1
    end do
  end subroutine

  subroutine sampling_priors(&
    pri,& 
!     $      Prior_ParaElements, Prior_ParaModules, Prior_Inits, 
!     $      ID_Prior_ParaElements, ID_Prior_ParaModules, 
!     $      ID_Prior_Inits,
    priG0,&
!     $      PriorG0_ParaElements, PriorG0_ParaModules, 
!     $      PriorG0_Inits,
!     $      max_num_Para1, max_num_Para2, num_Para1, num_Para2,
    sv,&
!     $      max_num_Variable, num_Variable,
!     $      max_num_Reg_Elements, max_num_Reg_Modules,
    mdl,&
!     $      num_Reg_Elements, num_Reg_Modules,
    ix,&
    par,&
!     $      Particle_ParaElements, Particle_ParaModules,
!     $      Particle_Inits, nsave_Particle,
!     $      List_Reg_Elements, List_Reg_Modules, strModules,
    method&
      )
    implicit real*8(a-h,o-z)
    implicit integer(i,j,k,l,m,n)
    type(T_Pri) :: pri, priG0
    type(T_SV)  :: sv
    type(T_Mdl) :: mdl
    integer     :: ix
    type(T_Par) :: par
    integer     :: method
    
    character(*), parameter :: tag_base = 'Baseline'

    if (method.eq.1) then
      w1 = 0.8d0
      s  = 0.01d0
    else if (method.eq.2) then
      w1 = 1.0d0
      s  = 0.0d0
    endif
    w2 = 1.0d0 - w1

!$omp parallel firstprivate(k,ix), private(id)
!$omp do
    do i = 1, pri%num_Para1
      k = pri%ID_Prior_ParaElements(i)
      do j = 1, mdl%num_Reg_Elements
        if (mdl%List_Reg_Elements(j,1) .eq. k) then
      10 continue
         pri%Prior_ParaElements(i,1)&
           = w1*par%Particle_ParaElements(j,1)&
           + w2*priG0%Prior_ParaElements(i,1)& 
           + s*dsqrt(priG0%Prior_ParaElements(i,3))*rgauss(ix)
         if (runif(ix) .ge. 0.9d0 .and. method.eq.1) then
           pri%Prior_ParaElements(i,1) = 0.1d0 + runif(ix)*1.0d0  
         endif
         if (pri%Prior_ParaElements(i,1) .lt. 0.0d0) goto 10
           pri%Prior_ParaElements(i,2)&
             = w1*par%Particle_ParaElements(j,2)&
             + w2*priG0%Prior_ParaElements(i,2)& 
             + s*dsqrt(priG0%Prior_ParaElements(i,4))*rgauss(ix)
         if( pri%Prior_ParaElements(i,2).lt. 0.0d0 ) goto 10     
         goto 1
        endif
      enddo
    1 continue
    enddo
!$omp end do    

!$omp do
    do i = 1, pri%num_Para2
      k = pri%ID_Prior_ParaModules(i)
      do j = 1, mdl%num_Reg_Modules
        if (mdl%List_Reg_Modules(j,3) .EQ. k) then
      20 continue    
          pri%Prior_ParaModules(i,1)&
            = w1*par%Particle_ParaModules(j)&
            + w2*priG0%Prior_ParaModules(i,1)&
            + s * dsqrt( priG0%Prior_ParaModules(i,2) )* rgauss(ix)
          if (pri%Prior_ParaModules(i,1) .lt. 0.0d0) goto 20     
          if (mdl%strModules(j,2) .EQ. tag_base)&
            pri%Prior_ParaModules(i,1) = 0.0d0    
            pri%Prior_ParaModules(i,2) = priG0%Prior_ParaModules(i,2)
          goto 2
        endif
      enddo
    2 continue
    enddo
!$omp end do    

!$omp do
    do i = 1, sv%num_Variable
!$    id = omp_get_thread_num()    
      k = pri%ID_Prior_Inits(i)
    30 continue    
      pri%Prior_Inits(i,1)&
        = w1*par%Particle_Inits(k)&
        + w2*priG0%Prior_Inits(i,1)& 
        + s*dsqrt(priG0%Prior_Inits(i,2))*rgauss(ix)
      if (pri%Prior_Inits(i,1) .lt. 0.0d0) goto 30     
      pri%Prior_Inits(i,2) = priG0%Prior_Inits(i,2)
    enddo
!$omp end do    
!$omp end parallel
  end subroutine

  subroutine Save_Particle(par, sv, mdl, num_overlap_DP, str)
    implicit real*8 (a-h, o-z)
    implicit integer (i-n)
    type (T_ParVec) :: par
    type (T_Mdl)    :: mdl
    type (T_SV)     :: sv
    integer :: num_overlap_DP(nsave_Particle)
    character(len=50) :: fileName
    character(len=*) :: str
 
    fileName = 'demo_particle_element_1_'//trim( str )//'.log'      
    open( 5, file = trim(fileName) )
    do i = 1, par%num_Particle
       write( 5, '( i8, 10000(a1,f15.5 ))')&
            num_overlap_DP(i),&
            (tab, par%o(i)%Particle_ParaElements(j,1), j = 1, mdl%num_Reg_Elements)
    end do
    close(5)
    
    fileName = 'demo_particle_element_2_'//trim( str )//'.log'
    open( 5, file = trim(fileName) )
    do i = 1, par%num_Particle
       write( 5, '( i8, 10000( a1, f15.5 ) )')&
       num_overlap_DP(i),& 
       (tab, par%o(i)%Particle_ParaElements(j, 2),&
       j = 1, mdl%num_Reg_Elements)
    end do
    close(5)

    fileName = 'demo_particle_module_'//trim(str)//'.log'
    open( 5, file = trim(fileName) )
    do i = 1, par%num_Particle
       write( 5, '( i8, 10000( a1, f15.5 ) )')& 
            num_overlap_DP(i),& 
            ( tab, par%o(i)%Particle_ParaModules(j),& 
            j = 1, mdl%num_Reg_Modules)
    end do
    close(5)

    fileName = 'demo_particle_inits_'//trim( str )//'.log'
    open( 5, file = trim(fileName) )
    do i = 1, par%num_Particle
       write( 5, '( i8, 10000( a1, f15.5 ) )' )&
            num_overlap_DP(i),&
            ( tab, par%o(i)%Particle_Inits(j),&
            j = 1, sv%num_Variable )
    end do
    close(5)
  end subroutine

  subroutine Set_Priors2Model(mdl, pri, sv, iprint)
    implicit real*8(a-h,o-z)
    implicit integer (i-n)

    type (T_Mdl) :: mdl
    type (T_Pri) :: pri
    type (T_SV)  :: sv
    integer      :: iprint

    if( iprint .EQ. 1 ) then
      write(*,*) 
      write(*,'(a)') '-@[Set_Priors2Model] ',&
        'Overwrite reaction kinetics on [ReactionModules] ',&
        'by the specified prior mean'
    endif

    do i = 1, pri%num_Para2
      k = pri%ID_Prior_ParaModules(i)
      if( iprint .EQ.1 ) write(*,'(a,a1,i3)') 'Prior@Modules', tab, k
      do j = 1, mdl%num_Reg_Modules
      if( mdl%List_Reg_Modules(j,3) .EQ. k ) then
      mdl%Para_Reg_Modules(j,1) = pri%Prior_ParaModules(i,1)
      iv = mdl%List_Reg_Modules(j,2)
      if( iprint .EQ. 1 ) write(*,'(a,a1,a,a1,a,a1,a)')&
           'Replace', tab, mdl%strModules(j,2), tab,& 
           '@Product', tab, sv%Name_Variable(iv)
      endif 
      enddo
    enddo

    if (iprint .EQ. 1) then
      write(*,*) 
      write(*,'(a)')&
          '-@[Set_Priors2Model] Overwrite reaction kinetics on ',&
          '[ReactionElements] by the specified prior mean'
    end if

    do i = 1, pri%num_Para1
      k = pri%ID_Prior_ParaElements(i)
      if( iprint .EQ.1 ) write(*,'(a,a1,i3)') 'Prior@Elements', tab, k
      do j = 1, mdl%num_Reg_Elements
      if( mdl%List_Reg_Elements(j,1) .EQ. k ) then
        mdl%Para_Reg_Elements(j,1) = pri%Prior_ParaElements(i,1)
        mdl%Para_Reg_Elements(j,2) = pri%Prior_ParaElements(i,2)      
        if( iprint .EQ. 1 ) write(*,'(10(a,a1))')& 
           'Replace', tab, mdl%strElements(j,3), tab,&
           '@Product', tab, mdl%strElements(j,1), tab,&
           '@Reactant',  tab, mdl%strElements(j,4)
        endif
      end do
    end do      

    if( iprint .EQ. 1 ) then
        write(*,*) 
        write(*,'(a)') '-@[Set_Priors2Model] ',&
          'Overwrite Initial States',&
          'by the specified prior mean'
    end if
      
    do i = 1, sv%num_Variable
      k = pri%ID_Prior_Inits(i)
      sv%StateVariable(k,1) = pri%Prior_Inits(i,1)
      if( iprint .EQ. 1 ) write(*,'(a,a1,a,a1,f10.5)')&
           'Replace', tab, sv%Name_Variable(k), tab,&
           pri%Prior_Inits(i,1)
    end do 
  end subroutine

  subroutine Save_Model2Particle(mdl, sv, par)
    implicit real*8(a-h,o-z)
    implicit integer (i-n)
    type (T_Par) :: par
    type (T_Mdl) :: mdl
    type (T_SV)  :: sv

    do i = 1, mdl%num_Reg_Elements
      do j = 1, 2
        par%Particle_ParaElements(i,j) = mdl%Para_Reg_Elements(i,j)
      end do
    end do

    do i = 1, mdl%num_Reg_Modules
      par%Particle_ParaModules(i) = mdl%Para_Reg_Modules(i,1)
    end do

    do i = 1, sv%num_Variable
      par%Particle_Inits(i) = sv%StateVariable(i,1)
    end do
  end subroutine 

  subroutine Save_Prior2Particle(pri, mdl, sv, par)
    type (T_Pri) :: pri
    type (T_Par) :: par
    type (T_Mdl) :: mdl
    type (T_SV)  :: sv
    integer      :: k, i, ii, j

!$omp parallel private(k)
!$omp do
    do i = 1, pri%num_Para1
      k = pri%ID_Prior_ParaElements(i)
      do ii = 1, mdl%num_Reg_Elements
        if (mdl%List_Reg_Elements(ii,1) .EQ. k) then
          do j = 1, 2
            par%Particle_ParaElements(ii,j) = pri%Prior_ParaElements(i,j)
          end do
        end if
      end do
    end do
!$omp end do

!$omp do
    do i = 1, pri%num_Para2
      k = pri%ID_Prior_ParaModules(i)
      do j = 1, mdl%num_Reg_Modules
        if (mdl%List_Reg_Modules(j,3) .EQ. k) then      
          par%Particle_ParaModules(j) = pri%Prior_ParaModules(i,1)
        end if
      end do
    end do
!$omp end do      

!$omp do
    do i = 1, sv%num_Variable
      k = pri%ID_Prior_Inits(i)
      par%Particle_Inits(k) = pri%Prior_Inits(i,1)
    end do
!$omp end do
!$omp end parallel
  end subroutine 

  subroutine Write_Particle&
     (par, mdl, sv, time_interval, pri&
     ,doRun, tag, includeNoise, ix, Variance_System, Variance_Measure)
    type (T_Par) :: par
    type (T_Mdl) :: mdl
    type (T_SV)  :: sv
    real(8) :: time_interval
    type (T_Pri) :: pri
    logical :: doRun
    character(*) :: tag
    integer :: includeNoise, ix, i, j
    real(8) :: Variance_System(max_num_Variable)
    real(8) :: Variance_Measure( max_num_Variable )
!$omp parallel 
!$omp do
    do i = 1, mdl%num_Reg_Elements
      do j = 1,2
        mdl%Para_Reg_Elements(i,j) = par%Particle_ParaElements(i,j)
      enddo
    enddo
!$omp end do
!$omp do
    do i = 1, mdl%num_Reg_Modules
      mdl%Para_Reg_Modules(i,1) = par%Particle_ParaModules(i)
    end do
!$omp end do  
    if (doRun) then
!$omp do
      do i = 1, sv%num_Variable
        sv%StateVariable(i,1) = par%Particle_Inits(i)
      end do
!$omp end do  
    end if 
!$omp end parallel 
    if (doRun) then
      call Run_BioSimulator_version2&
        (mdl, sv, time_interval, includeNoise, ix, Variance_System)
    end if 

     call Wrap_Output_Reactions&
      (mdl, sv, time_interval, pri, Variance_System, Variance_Measure&
      ,trim(mdl%suffix)//'_'// tag //'.emdl')

    call Wrap_Output_BioSimulator&
      (sv, 'simulation_'// tag  //'.log')
  end subroutine

  ! test whether map sol. is updated, and where if done.
  ! input:
  !   value_max_dpm: ...
  !   value_max: the current (global) maxium likelihood
  !   Nid: my rank
  ! output:
  !   valueLik_max: ...
  !   (return value): the node number (>=0) yields the new map or -1 if not updated.
  integer function itest_map_update(value_max_dpm, valueLik_max, Nid)
    use mpi
    real(8) :: value_max_dpm, valueLik_max
    real(8) :: val_node_(2), val_node(2), valueLik_max_
    integer :: Nid, ierr

    val_node_(1)  = value_max_dpm
    val_node_(2)  = Nid
    valueLik_max_ = valueLik_max
    CALL MPI_AllReduce&
      (val_node_, val_node, 1, MPI_2DOUBLE_PRECISION, MPI_MAXLOC, MPI_COMM_WORLD, ierr)
    if (val_node(1) .gt. valueLik_max) then
      valueLik_max     = val_node(1)
      itest_map_update = val_node(2)
    end if
  end function

  subroutine dGauss_Particles_from_Center(prs, par, parC, mdl, sv, pri)
    use omp_lib
    type (T_ParVec) :: par
    type (T_Par)    :: parC
    type (T_SV)     :: sv
    type (T_Mdl)    :: mdl
    type (T_Pri)    :: pri
    real(8)         :: prs(:)
    integer :: i, j, k, ii, id
    real(8) :: x1,x2,d1,d2,ss, u1, u2, sum

    double precision, external :: dGauss_Univ

!$omp parallel private(k,u1,u2,x1,x2,d1,d2,ss)
!$  id = omp_get_thread_num()
    ss = 0.1d0
!$omp do
    do i = 1, pri%num_Para1
      k = pri%ID_Prior_ParaElements(i)
      do ii = 1, mdl%num_Reg_Elements
        if (mdl%List_Reg_Elements(ii,1) .EQ. k) then
          u1 = parC%Particle_ParaElements(ii,1)
          u2 = parC%Particle_ParaElements(ii,2)
          do j = 1, nsave_Particle
            x1 = par%o(j)%Particle_ParaElements(ii,1)
            !x2 = par%o(j)%Particle_ParaElements(ii,1)!bug?
            x2 = par%o(j)%Particle_ParaElements(ii,2)
            d1 = dGauss_Univ(x1, u1, ss)
            !d2 = dGauss_Univ(x2, u2, ss)
            prs(j) = log(d1) !+ log(d2)
          end do
          exit 
        end if
      end do
    end do
!$omp end do
!$omp do
    do i = 1, pri%num_Para2
      k = pri%ID_Prior_ParaModules(i)
      do ii = 1, mdl%num_Reg_Modules
        if( mdl%List_Reg_Modules(ii,1) .EQ. k ) then
          u1 = parC%Particle_ParaModules(ii)
          do j = 1, nsave_Particle
            x1 = par%o(j)%Particle_ParaModules(ii)
            d1 = dGauss_Univ( x1, u1, ss )
            prs(j) = prs(j) + log(d1)
          end do
          exit
        endif
      end do
    end do
!$omp end do
!$omp do
    do i = 1, sv%num_Variable
      k = pri%ID_Prior_Inits(i)    
      u1 = parC%Particle_Inits(k)
      do j = 1, nsave_Particle
        x1 = par%o(j)%Particle_Inits(k)
        d1 = dGauss_Univ( x1, u1, ss )
        prs(j) = prs(j) + log(d1)
      end do
    end do
!$omp end do
!$omp end parallel

      sum = 0.0d0 
!$omp parallel do reduction(+:sum)
      do i = 1, nsave_Particle
        sum = sum + dexp(prs(i))
      end do
!$omp end parallel do
!$omp parallel do
      do i = 1, nsave_Particle
        prs(i) = dexp(prs(i))/sum
      end do
!$omp end parallel do
  end subroutine

  subroutine Save_Snapshot(outfile, dat, mdl, sv, time_interval, Variance_System)
    character(*) :: outfile
    type (T_Dat) :: dat
    type (T_Mdl) :: mdl
    type (T_SV)  :: sv
    real(8)      :: time_interval
    real(8)      :: Variance_System(:)
    open(1234, file=outfile, form='BINARY', convert='BIG_ENDIAN')
      write(1234) dat
      write(1234) mdl
      write(1234) sv
      write(1234) time_interval
      write(1234) Variance_System
    close(1234)
  end subroutine

  subroutine Load_Snapshot(infile, dat, mdl, sv, time_interval, Variance_System)
    character(*) :: infile
    type (T_Dat) :: dat
    type (T_Mdl) :: mdl
    type (T_SV)  :: sv
    real(8)      :: time_interval
    real(8)      :: Variance_System(:)
    open(1234, file=infile, form='BINARY', convert='BIG_ENDIAN')
      read(1234) dat
      read(1234) mdl
      read(1234) sv
      read(1234) time_interval
      read(1234) Variance_System
    close(1234)
  end subroutine
end module
