!-------------------------------------------------------------------------------
! LiSDAS v2.7
!
! Copyright 2010
!   YOSHIDA, Ryo,
!   SAITO, Masaya,
!   NAGAO, Hiromichi,
!   NAKANO, Shin'ya,
!   HIGUCHI, Tomoyuki,
!   The Institute of Statistical Mathematics
!
! Dual licensed under the MIT or GPL Version 3 licenses.
! See 'LICENSE.MIT' for the definition of the MIT license, 
! and 'COPYING.GPL' for the definition of the GPL license.
!-------------------------------------------------------------------------------
! File name: LiSDAS-2-7-mpi.f
!
! Code name: LiSDAS_2_7
! 
! Package: LiSDAS
!
! Language: Fortran77/95, OpenMP, MPI
!
! Description: Parameter learning of generalized state space models using 
!              DPM priors with Bayesian expert approach 
!              (a hybrid parallel implementation using OpenMP and MPI)
!
! Author: Ryo Yoshida
!           The Institue of Statistical Mathematics, 
!           Research Organization of Information and Systems
!           email; yoshidar@ism.ac.jp
!           url; http://daweb.ism.ac.jp/~yoshidar/
!
! Last update:      21 November 2011
!
! References:            
!   [1] Yoshida,R, Nagasaki,M, Yamaguchi,R, Imoto,S, Miyano,S, Higuchi,T,  
!       Bayesian learning of biological pathways on genomic data assimilation,
!       Bioinformatics, 24(22):2592-2601, 2008. 
!
!   [2] Yoshida,R, Saito,M.M, Nagao,H, Higuchi,T, Bayesian experts in exploring
!       reaction kinetics of transcription circuits, Bioinformatics, 
!       26(18):i589-595, 2010. 
!-------------------------------------------------------------------------------
      program LiSDAS_2_7
!     --- OpenMP module ---
!$    use omp_lib
      use M_LiSDAS 

      implicit real*8 ( a-h, o-z )

!     --- MPI code ---
      Include 'mpif.h'

!     ---  Array allocation ---
!     / Covariance matrix and mean vector of parameters in "Reaction Elements", 
!       sequentially updated while raising the number of particles
      dimension Ave_ParaElements(max_num_Reg_Elements)
      dimension Cov_ParaElements(max_num_Reg_Elements, max_num_Reg_Elements)

!     / Covariance matrix and mean vector of parameters in "Reaction Modules", 
!       sequentially updated while raising the number of particles
      dimension Ave_ParaModules(max_num_Reg_Modules)
      dimension Cov_ParaModules(max_num_Reg_Modules, max_num_Reg_Modules)

!     / Time course data and state variables
      type (T_Dat) :: dat
      type (T_SV)  :: sv

      ! Oops slightly differ from T_Par ... 
      dimension st_map( max_num_Variable, 10000 )
      dimension pelement_map( max_num_Reg_Elements, max_tmp)
      dimension pmodule_map( max_num_Reg_Modules, max_tmp)

!     / Log-likelihood of each particle
      dimension valueLik( max_num_Particle )

!/ Matrices defining structure and reaction kinetics of model
      type (T_Mdl) :: mdl

!/ Maintained particles for the parameters in "Reaction Elements", 
!  "Reaction Modules", and "Initial States"
      type (T_ParVec) :: elite

!/ [Point mass functions of DPM] Maintained particles for the parameters in 
!  "Reaction Elements", "Reaction Modules", and "Initial States"
      type (T_ParVec) :: dp

      dimension list_m( max_num_Para2 )
      dimension list_e( max_num_Para1 )
      dimension list_i( max_num_Variable )
      dimension list_mm( max_num_Para2, max_tmp )
      dimension list_ee( max_num_Para1, max_tmp ) 
      dimension list_ii( max_num_Variable, max_tmp )      
      dimension grad_m( nsave_Particle, max_num_Para2 )
      dimension grad_e( nsave_Particle, max_num_Para1 ) 
      dimension grad_i( nsave_Particle, max_num_Variable )            

!      / Log-likelihoods, normalized weights for the resampling,
!       and IDs of resampled particles, and sampling counts of point mass functions of DPM
      dimension saveLik( nsave_Particle )
      dimension weight4resample( nsave_Particle )
      dimension prs( nsave_Particle )
      dimension prsPlusOne( nsave_Particle + 1 )      
      dimension ismp_Particle( nsave_Particle )
      dimension num_overlap_DP( nsave_Particle )

!      / Prior specification matrices for "Reaction Elements", 
!       "Reaction Modules", and "Initial States"
      type (T_Pri) :: pri

!     / Parameters of base measure of DP for "Reaction Elements", 
!       "Reaction Modules", and "Initial States" 
      type (T_Pri) :: priG0

!      / Measurement and system noise variances 
      dimension Variance_Measure( max_num_Variable )
      dimension Variance_System( max_num_Variable )

!      / Summary statistics for experimental data
      dimension ave_varData( max_num_Variable, 2 )      ! Mean and variance of time course data
      dimension value_maxminData( max_num_Variable, 2 ) ! Max and Min of time course data
      dimension loc_maxmin( max_num_Variable, 2 )       ! Time points in which the max or min data value appears

!     / Upper/Lower bounds of varaibles for time course data and model       
      dimension bound_data( max_num_Variable, 2 )
      dimension bound_model( max_num_Variable, 2 )      

!      / Variables involved in Gaussian kernel regression
      dimension CoeffGaussKernel( max_nbasisGauss, max_num_Variable )
      dimension gradient( max_num_Variable, max_time_StateVariable )
      dimension centerKernel( max_time_Data )
      integer   bn_gradient( max_num_Variable, max_time_StateVariable )      
      dimension resid( max_num_Variable, max_time_Data )
      dimension RefState( max_num_Variable, max_time_StateVariable )      

!      / Duration times with increasing/decreasing phase
      dimension durTimeI( max_num_Variable, nDecompTime, 2 )    
      dimension durTimeD( max_num_Variable, nDecompTime, 2 )
      dimension nDurTimeI( max_num_Variable )
      dimension nDurTimeD( max_num_Variable )
      dimension durMaxTimeI( max_num_Variable, 2 )
      dimension durMaxTimeD( max_num_Variable, 2 )
      dimension degrad_RNA( max_num_Variable )
      dimension synthesis_RNA( max_num_Variable )

!     / Molecular information and regulatory relationships
      parameter ( max_num_transcript = 1000 )
      dimension imRNA( max_num_Variable )
      dimension itranslation( max_num_Variable )
      dimension icomplex( max_num_Variable )
      dimension itranscript( max_num_transcript, 3 )
      dimension switch( max_num_transcript ) 
      dimension List_Reg_Elements_switch( max_num_transcript )     

!     / Gaussian kernel regression
      dimension bandKernel( max_num_Variable )      

      character Output_File * 30 ! Name of output file 

      character str * 15
      character str2 * 15      

!      / Computational times 
      integer*4 begin_clock
      integer*4 end_clock
      dimension CompTime( max_num_Particle )

!     / Additional parameters relavant to the DPM 
!      parameter ( max_num_Particle_DPM = 200 )
      dimension weight_DPM( max_num_Particle_DPM )
      dimension pr_DPM( max_num_Particle_DPM )      

!     / Range of (Upper/Lower) values taken for unobserved regulatory variables
      dimension RangeHidden(2)

!     / Parameters related to graphical structure of model
      parameter ( max_num_edge = 1000 )

!     @ (1) Child, (2) Parent     
      dimension network( max_num_edge, 2 )
!     @ Attributes of 'network': (1) Name of child, (2) Name of parent, (3) Reaction, 
!       (4) Positive or Negative, (5) Transcription Activotor or Inhibitor      
      character network_attribute( max_num_edge, 5 ) * 30

!     @ Reduced graph for mRNA variable: (1) Child, (2) Parent, (3) Positive or Negative, 
!       (4) Transcription Activotor or Inhibitor            
      dimension network_rna( max_num_edge, 4 )      

!     / Summary statistics of time course data
      dimension cov_data( max_num_Variable, max_num_Variable )
      dimension ave_data( max_num_Variable )

!      / Some statistics evaluating sampling efficiency
      dimension itr_replace_map(100000)
      dimension itr_replace_dpm(100000)

      dimension discountRateTranscpript(2) 

! --- MPI start ---
      Call MPI_Init(ierr)
      CALL MPI_Comm_size(MPI_COMM_WORLD, Nproc, ierr)
      CALL MPI_Comm_rank(MPI_COMM_WORLD, Nid, ierr)      

! --- Display version information ---
      if( Nid. EQ. 0 ) then
      write(*,*)
      write( *, '(a)' ) 'Package: LiSDAS'
      write( *, '(a)' ) 'Code: LiSDAS_2_7'
      write( *, '(a)' ) 'Author: Ryo Yoshida'
      write( *, '(a)' ) 'Last update: 21-Nov-2011'

      write(*,*) 
      write( *, '(a,i5)' ) 'Use MPI:'
      write( *, '(a,i5)' ) 'Use number of PE:', Nproc 
      endif

! ---  Initialization ---
      itr = 0                   ! Iteration number for the parameter search
      num_savefile = 0
      num_saveSimul = 0
      sumLik = 0.0d0                  ! Sum of likelihood scores
      locSaveLik = 1                  ! Address of [saveLike], top simulators 
      sumCompTime = 0.0d0            ! Total computational time
      isample_G0 = 0            ! Sampling indicator for base measure of DP 
      num_edge = 0              ! Number of edges in graphical structure of input model 
      num_edge_rna = 0          ! Number of edges in reduced mRNA graph of input model
      icount_replace_base = 0
      input_zero_baseline = 1
!
      num_replace_map = 0
      num_replace_dpm = 0
!      
      discountRateTranscpript(1) = 1.3d0
      discountRateTranscpript(2) = 0.5d0       

! --- Procedure specification 'procedure.set' ---
!
      open( 5, file = 'procedure.set' )
      read( 5, * )
      read( 5, * )
      read( 5, * ) iprint_process      ! {0,1} Output descriptions abount computational process      
      read( 5, * ) iDoSaveSimul            ! {0,1} Save a history of simulations to 'save_simulation_*.log'
      read( 5, * ) numDoSaveSimul      ! Number of simulations to be saved
      read( 5, * ) iDoSaveCovar            ! {0,1}      Save covariance matrix of the generate particles
      read( 5, * ) iDoSaveParticle      ! {0,1}      Keep in computational process a set of survived particles and their likelihoods,
!                                      and reuse them to create next generations of particles
      read( 5, * ) iDoDPM                  ! {0,1}      Do the adaptive DPM sampling
      read( 5, * ) numDoDPM                  ! Max number of point mass functions in DPM analysis
      read( 5, * ) iDoStrSearch            ! {0,1}      Do the search for circuit structureres of biochemical reactions
      read( 5, * ) includeNoise            ! {0,1}      Do the stochastic simulation during parameter search (inclusion of system noise)  
      read( 5, * ) iDoClock           ! {0,1} Do CPU time evaluation
      read( 5, * ) bwkernel         ! Bandwidth parameter for Gaussian kernel regression
      read( 5, * ) nbw              ! Number of bandwidths for the choise based on LOOCV
      read( 5, * ) shrinkPara       ! Ridge parameter for nonparametric regression
      read( 5, * ) iDoSampleModule  ! Method of drawing particles for parameters in "Module" attribute 
      read( 5, * ) iDosampleElement ! Method of drawing particles for parameters in "Element" attribute
      read( 5, * ) iDoSampleInits   ! Method of drawing particles for initial state   
      read( 5, * ) iPrint           ! {0,1} Print the computational process
!
      read( 5, * ) iUseDataInits    ! {0,1} Use observation at the first time point as the initial state (for initial simulation)
      read( 5, * ) iDoRangeHidden   ! {0,1} Impose range of values taken for hidden regulatory variables
      read( 5, * ) RangeHidden(1)   ! Upper value for the hidden variables
      read( 5, * ) RangeHidden(2)      ! Lower value for the hidden variables
      read( 5, * ) num_Particle_DPM ! Maximum number of point masses to be saved for DPM sampling
      read( 5, * ) gamma            ! Concentration parameter of Dirichlet process
      close( 5 )

      !/Setting sizes of pools of particles      
      dp%num_Particle    = num_Particle_DPM ! But, the code seems to assume = nsave_Particle
      elite%num_Particle = nsave_Particle

!     (Print progress)
!     ------------------------------------------------------------------
      if( Nid. EQ. 0 ) then
      if( iPrint. EQ. 1 ) then
!
      write(*,*)
      write(*,'(a)') 
     $ '-@Get experimental parameters from [procedure.set]'
!
      if( iDoSaveSimul.EQ.1 ) then
        write(*,'(a)')
     $ 'TRUE: Simulation trajectories stored in [save_simulation_*.log]'
        write(*,'(a8,i3,a3,i3)') '@ "*" =', 2, '-', numDoSaveSimul      
      endif
!
      if( iDoStrSearch.EQ.1)
     $  write(*,'(a)') 
     $  'TRUE: Run a structural search program'
!
      if( includeNoise.EQ.1 )
     $  write(*,'(a)') 
     $  'TRUE: Include system noise'
!
      if( iDoClock.EQ.1 )
     $  write(*,'(a)') 
     $  'TRUE: Run time evaluation'
!
      endif
      endif
!     ------------------------------------------------------------------      
!
!
! --- Get experimental parameters involving random number generators from 'RAND.set' ---
!
      open( 5, file = 'RAND.set' )
      read( 5, * ) ix                        ! Initial seed for random numbers
      read( 5, * ) num_Particle            ! Total number of particles used in the search
      close( 5 )
!
!
!     (Print progress)
!     ------------------------------------------------------------------
      if( Nid. EQ. 0 ) then
      if( iPrint. EQ. 1 ) then
      write(*,'(a)')
     $  '-@Get number of particles from [RAND.set]'
      write(*,'(a,i)') 'Number of particles =', num_Particle
      endif
      endif
!     ------------------------------------------------------------------     

! --- Get a mdl%suffix of input/output file names from 'mdl%suffix.set' ---
      call Get_Suffix( mdl%suffix )

!     (Print progress)
!     ------------------------------------------------------------------
      if( Nid. EQ. 0 ) then
      if( iPrint. EQ. 1 ) then      
      write(*,*)
      write(*,'(a)') 
     $ '-@[Get_Suffix] Get a file mdl%suffix from [mdl%suffix.set]'
      write(*,*)      
      write(*,'(a,a)') 'Analysis identifier: ', mdl%suffix
      write(*,'(a)') 'Names of output files: ' 
      write(*,'(a)') trim(mdl%suffix)//'_input.emdl'
      write(*,'(a)') trim(mdl%suffix)//'.emdl'
      endif
      endif
!     ------------------------------------------------------------------      
! --- Get a list of reaction functions, its structure and prior parameters from '[mdl%suffix].mdl' ---
      call Wrap_Get_Reactions(mdl, sv, time_interval, pri)
      priG0%num_Para1 = pri%num_Para1
      priG0%num_Para2 = pri%num_Para2

! --- Get time series data from '[mdl%suffix].dat' ---
      call Get_Data(
!            -- Observed time course data
     $            dat%Data, dat%time_Data, sv%num_Variable, 
     $            max_time_Data, max_num_Variable, 
!            -- Indicators for observed time points and missing variables
     $            dat%ID_Time_Data, dat%ID_MissVariable, 
!            -- Others
     $            mdl%suffix, sv%Name_Variable 
     $      )

!     (Print progress)
!     ------------------------------------------------------------------
      if (Nid. EQ. 0 .AND. iPrint. EQ. 1) then
        write(*,*)
        write(*,'(a)') 
     $    '-@[Get_Data] Get time course observations from [*.dat]'
        write(*,'(a,a)') 'Accessible: ', trim( mdl%suffix )//'_data.log'
      endif
!     ------------------------------------------------------------------         
!
!      call renumbering_reactions(
!       
!            -- Reaction functions 
!     $            mdl%List_Reg_Elements, mdl%Para_Reg_Elements,
!     $            mdl%List_Reg_Modules, mdl%Para_Reg_Modules,
!     $            mdl%num_Reg_Elements, mdl%num_Reg_Modules, 
!     $            max_num_Reg_Elements, max_num_Reg_Modules, max_tmp, 
!
!            -- Variables
!     $            sv%ID_Variable, sv%Name_Variable, sv%num_Variable, 
!     $            max_num_Variable,
!      
!            -- Prior distributions        
!     $            pri%Prior_ParaElements, pri%ID_Prior_ParaElements, 
!     $            pri%Prior_ParaModules, pri%ID_Prior_ParaModules,
!     $            pri%Prior_Inits, pri%ID_Prior_Inits, 
!     $            pri%num_Para1, pri%num_Para2, max_num_Para1, max_num_Para2,
!
!     $            sv%StateVariable, max_time_StateVariable, sv%time_StateVariable,
!     $            dat%Data, max_time_Data, dat%time_Data,     
!
!            -- Variances
!     $            Variance_System, Variance_Measure,
!     $            mdl%strElements, mdl%strModules, mdl%strParaElements, mdl%strParaModules
!      
!     $ )

! --- Output the input model to '[mdl%suffix]_input.emdl' ---
      if( Nid. EQ. 0 ) then
        call Wrap_Output_Reactions
     $    (mdl, sv, time_interval, pri, Variance_System, Variance_Measure
     $    ,trim(mdl%suffix)//'_input.emdl')
      endif

!     (Print progress)
!     ------------------------------------------------------------------
      if( Nid. EQ. 0 ) then
      if( iPrint. EQ. 1 ) then      
      write(*,*)
      write(*,'(a)') 
     $ '-@[Get_Reactions]
     $ Get a reaction set from LiSDAS definition file [*.mdl]' 
      write(*,'(a,a)') 'Accessible: ', trim(mdl%suffix)//'_input.emdl'
      endif
      endif
!     ------------------------------------------------------------------      
!
!
! --- Get variables associated to proteins encoded by mRNAs  ---
!
      call relation_translation(
!      
     $      mdl%List_Reg_Modules, mdl%List_Reg_Elements, mdl%strModules,
     $      mdl%num_Reg_Modules, mdl%num_Reg_Elements,
     $      max_num_Reg_Modules, max_num_Reg_Elements, max_tmp, 
     $      itranslation, sv%num_Variable, max_num_Variable,
     $      sv%Name_Variable, sv%ID_Variable, imRNA
!
     $  )
!
      if( Nid. EQ. 0 ) then
      Output_File = 'rna_protein.log'
      call OutputVector_Integer( 
     $      imRNA, sv%num_Variable, max_num_Variable, Output_File
     $  )
      endif
!
!
! --- Get variables associated to protein complex ---
!
      call relation_complex(
!
     $      mdl%Para_Reg_Modules,      
     $      mdl%List_Reg_Modules, mdl%List_Reg_Elements, mdl%strModules,
     $      mdl%num_Reg_Modules, mdl%num_Reg_Elements,
     $      max_num_Reg_Modules, max_num_Reg_Elements, max_tmp, 
     $      icomplex, sv%num_Variable, max_num_Variable,
     $      sv%Name_Variable, sv%ID_Variable
!
     $  )
!     
!
      call covariance_matrix( cov_data, ave_data, dat%Data, 
     $      dat%time_Data, sv%num_Variable, max_time_Data, max_num_Variable )
!
!
! --- Network structure ---
!
      call get_network(
!      
!            -- Reaction functions       
     $      mdl%List_Reg_Elements, mdl%Para_Reg_Elements,
     $            mdl%List_Reg_Modules, mdl%Para_Reg_Modules,
     $            mdl%num_Reg_Elements, mdl%num_Reg_Modules, 
     $            max_num_Reg_Elements, max_num_Reg_Modules, max_tmp,
     $      mdl%strModules, mdl%strElements,
!       -- Variables
     $            sv%ID_Variable, sv%Name_Variable, sv%num_Variable, 
     $            max_num_Variable, mdl%suffix, 
     $      itranslation, icomplex,
!       -- Network
     $      network, network_attribute, num_edge, max_num_edge,
     $      network_rna, num_edge_rna     
!
     $  )
!
!
! --- Get system and observational noise variances from '[mdl%suffix].mdl'---
!
      call Get_NoiseVariances(
!
!            -- Variances
     $            Variance_System, Variance_Measure, sv%num_Variable,
     $            max_num_Variable,
!
!            -- Others
     $            mdl%suffix
!
     $      )
!
!
!     (Print progress)
!     ------------------------------------------------------------------
      if( Nid. EQ. 0 ) then
      if( iPrint. EQ. 1 ) then      
      write(*,*)
      write(*,'(a)') 
     $  '-@[Get_NoiseVariances] Get noise variances from [*.mdl]' 
      endif
      endif
!     ------------------------------------------------------------------      
!
!
! --- Compute summary statistics of observed data ---
!
      call SummaryStatistics(
!      
     $          dat%Data, sv%num_Variable, dat%time_Data, 
     $          max_num_Variable, max_time_Data, 
     $          ave_varData, value_maxminData, 
     $      loc_maxmin
!     
     $      )
!
!
!     (Print progress)
!     ------------------------------------------------------------------
      if( Nid. EQ. 0 ) then
      if( iPrint. EQ. 1 ) then      
      write(*,*)
      write(*,'(a)') 
     $  '-@[SummaryStatistics]
     $  Summary statistics of time course data'
      write(*,'(a,a)') 
     $ 'Accessible: ', trim( mdl%suffix )//'_statistics.log'     
      endif
      endif
!     ------------------------------------------------------------------      
!
!
!     (Output summary statistics of observed data)
!     ------------------------------------------------------------------
      if( Nid. EQ. 0 ) then
      open(5, file = trim( mdl%suffix )//'_statistics.log' )
      write( 5,'(1000(a,a1))') 
     $  'NameOfVariable', tab, 
     $  'ID', tab, 
     $  'Observability', tab, 
     $  'Maximum', tab, 
     $  'Minimum', tab, 
     $  'Average', tab, 
     $  'Variance'
      do i = 1, sv%num_Variable
      write( 5,'( a, a1, i, a1, i, 1000(a1,f10.5) )')
     $      sv%Name_Variable(i), tab, 
     $  sv%ID_Variable(i), tab,
     $  dat%ID_MissVariable(i),  
     $      ( ( tab, value_maxminData(i,j) ), j=1,2 ),
     $  ( ( tab, ave_varData(i,j) ), j=1,2 )
      enddo
      close(5)
      endif
!     ------------------------------------------------------------------      
!
!
! --- Bandwidth selection for Gaussian kernel regression (LOOCV) ---
!
      do i = 1, sv%num_Variable
      bandKernel(i) = bwkernel
      enddo
!      
      iDoPrint = 1
      call Gauss_Kernel_Regression_BandSelect(
!      
     $            dat%Data, sv%StateVariable, sv%time_StateVariable, dat%time_Data, 
     $            sv%num_Variable, max_time_StateVariable, max_time_Data, 
     $            max_num_Variable,
     $            bandKernel, shrinkPara, nbw,
     $            dat%ID_MissVariable, dat%ID_Time_Data,
     $            CoeffGaussKernel, max_nbasisGauss, iDoPrint
!
     $      )
!
!
!     (Print progress)
!     ------------------------------------------------------------------
      if( Nid. EQ. 0 ) then
      if( iPrint. EQ. 1 ) then      
      write(*,*)
      write(*,'(a)')
     $  '-@[Gauss_Kernel_Regression_BandSelect]
     $  Get bandwidth parameters for Gaussian kernel regression'
      endif
      endif
!     ------------------------------------------------------------------      
!     
!
! --- Gaussian kernel regression ---
!
      iDoPrint = 1
      call Gauss_Kernel_Regression(
!
     $            dat%Data, sv%StateVariable, sv%time_StateVariable, dat%time_Data, 
     $            sv%num_Variable, max_time_StateVariable, max_time_Data, 
     $            max_num_Variable,
     $            bandKernel, shrinkPara, 
     $            dat%ID_MissVariable, dat%ID_Time_Data,
     $            CoeffGaussKernel, max_nbasisGauss, resid, iDoPrint
!
     $      )
!
!
!     (Print progress)
!     ------------------------------------------------------------------
      if( Nid. EQ. 0 ) then      
      if( iPrint. EQ. 1 ) then      
      write(*,*)
      write(*,'(a)') 
     $  '-@[Gauss_Kernel_Regression]
     $  Get an interpolation of data using Gaussian kernel regression'
      endif          
!
      if( iDoPrint. EQ. 1 ) then      
      write(*,*)
      write(*,'(a)')
     $  '-Gaussian kernel regression'
      write(*,'(a,a)')
     $  'Accessible: ', 'gauss_kernel.log'
      endif
      endif
!     ------------------------------------------------------------------
!
! --- Evaluate gradients of the fitted Gaussian kernel regressions ---
!
      iDoPrint = 1
      call Gauss_Kernel_Gradient(
!
!            --- Gradient vectors [gradient] of given basis functions are evaluated over 1:[timeAll]            
     $            gradient, sv%time_StateVariable, dat%time_Data, sv%num_Variable, 
     $            max_num_Variable, max_time_StateVariable, max_time_Data,
!
!            --- [band] Bandwidth (variance) of Gauss kernels, [centerKerne] Center of basis functions -> 1:[timeMeasure]        
     $            bandKernel, centerKernel,
!
!            --- [measure_variable] Observed variables in data, [measure_time] Observed time points in data
     $            dat%ID_MissVariable, dat%ID_Time_Data,
!
!            --- [coeff] Estimated coefficients (number of basis functions times number of variables)
     $            CoeffGaussKernel, max_nbasisGauss, nbasisGauss,
!
!            --- Output file indicator
     $            iDoPrint
!
     $      )
!
!
!$omp parallel 
!$omp do
      do i = 1, sv%num_Variable
      do j = 1, sv%time_StateVariable
      if( gradient(i,j). ge. 0.0d0 ) bn_gradient(i,j) = 1
      if( gradient(i,j). lt. 0.0d0 ) bn_gradient(i,j) = 0
      enddo
      enddo
!$omp enddo
!$omp end parallel      
!
!     (Print progress)
!     ------------------------------------------------------------------
      if( Nid. EQ. 0 ) then
      if( iPrint. EQ. 1 ) then      
      write(*,*)
      write(*,'(a)')
     $  '-@[Gauss_Kernel_Gradient]
     $  Get gradients of smoothed estimates of data points'
      endif          
!
      if( iDoPrint. EQ. 1 ) then      
      write(*,*)
      write(*,'(a)') 
     $  '-@[Gauss_Kernel_Gradient]
     $  Gradients of Gaussian kernel regression'
      write(*,'(a,a)') 
     $  'Accessible: ', 'gauss_kernel_gradient.log'     
      endif
      endif      
!     ------------------------------------------------------------------      
!
!
! --- Get duration times for each variables and increasing/decreasing phase according to the evaluated gradients ---
!
      iDoPrint = 1
      call Gauss_Kernel_Duration(
!
     $            durTimeI, durTimeD, nDurTimeI, nDurTimeD,
     $            gradient, max_num_Variable, sv%num_Variable, 
     $            max_time_StateVariable, sv%time_StateVariable,
     $            nDecompTime, dat%ID_MissVariable, iDoPrint
!
     $      )
!
!
!     (Print progress)
!     ------------------------------------------------------------------
      if( Nid. EQ. 0 ) then
!
      if( iPrint. EQ. 1 ) then      
      write(*,*)
      write(*,'(a)') 
     $  '-@[Gauss_Kernel_Duration]
     $  Get increasing/decreasing durations'
      endif          
!
      if( iDoPrint. EQ. 1 ) then      
      write(*,*)
      write(*,'(a)') 
     $  '-@[Gauss_Kernel_Duration]
     $  Decomposed times based on increasing/decreasing trends'
      write(*,'(a,a)') 
     $  'Accessible: ', 'gauss_kernel_duration.log'     
      endif
!
      endif      
!     ------------------------------------------------------------------      
!
!
! --- Get a longest duration with increasing/decreasing phase for each variable ---
!
      call Duration_Select_Max(
!
!            --- Max duration time for each variable
     $            durMaxTimeI, durMaxTimeD,
!           
!            --- Set of duration times in increasing and decreasing trends
     $            durTimeI, durTimeD, nDurTimeI, nDurTimeD, nDecompTime,
!
!            --- Others
     $            max_num_Variable, sv%num_Variable, dat%ID_MissVariable,
     $            max_time_StateVariable      
!
     $      )
!
!
!     (Print progress)
!     ------------------------------------------------------------------
      if( Nid. EQ. 0 ) then
      if( iPrint. EQ. 1 ) then      
      write(*,*)
      write(*,'(a)') 
     $  '-@[Duration_Select_Max] Get maximal durations'
      endif               
      endif
!     ------------------------------------------------------------------
!
!
! --- Compute some base measure parameters relevant to mRNA synthesis (rates of degradation and transcription) ---
!
!      */ [degrad_RNA] and [synthesis_RNA] are used as Bayesian experts in local reaction speeds 
!
      call Duration_SteadyState(
!
!            --- Rates of degradation and transcription
     $            degrad_RNA, synthesis_RNA, 
!
!            --- Time ranges (start and end) for increasing and decreasing phases
     $            durMaxTimeI, durMaxTimeD,
!
!            --- Missing time series indicator, time interval
     $            dat%ID_MissVariable, time_interval,
!
!            --- State variables (reference)
     $            sv%StateVariable, max_num_Variable, sv%num_Variable, 
     $            max_time_StateVariable, sv%time_StateVariable, 
!
!       --- Turning parameters
     $      discountRateTranscpript
     $      )
!
!
!     (Print progress)
!     ------------------------------------------------------------------
      if( Nid. EQ. 0 ) then
      if( iPrint. EQ. 1 ) then      
      write(*,*)
      write(*,'(a)') 
     $  '-@[Duration_SteadyState]
     $      Compute transcriptional parameters satisfying
     $      steady-state conditions'
      endif
      endif
!     ------------------------------------------------------------------      
!
!
! --- Reference profiles ---
!
      call Reference_Profiles(
!
     $      sv%StateVariable, max_num_Variable, sv%num_Variable, 
     $      max_time_StateVariable, sv%time_StateVariable, 
     $      sv%Name_Variable, dat%ID_MissVariable,
     $      network, network_attribute, itranslation, icomplex,
     $      num_edge, max_num_edge, ix, RefState
! 
     $  )
!
!
!
! --- Shift prior distribution on Reaction Modules ---
!
!      
      call Shift_Prior_Modules(
!      
     $      mdl%List_Reg_Modules, mdl%Para_Reg_Modules, max_num_Reg_Modules, 
     $      mdl%num_Reg_Modules, mdl%strModules, max_tmp, 
     $      pri%Prior_ParaModules, pri%ID_Prior_ParaModules, 
     $      max_num_Para2, pri%num_Para2, ix,
     $      degrad_RNA, synthesis_RNA, max_num_Variable, sv%num_Variable,
     $      sv%Name_Variable, sv%ID_Variable, dat%ID_MissVariable,
     $      itranslation, icomplex, value_maxminData, RangeHidden,
     $      input_zero_baseline
!
     $ )
!
!
! --- Shift prior distribution on Initial States ---
!
      call Shift_Prior_Inits(
!      
     $      pri%Prior_Inits, pri%ID_Prior_Inits,
     $      itranslation, icomplex, value_maxminData, 
     $      ave_varData, RangeHidden,
     $      sv%Name_Variable, max_num_Variable, sv%num_Variable,
     $      dat%Data, max_time_Data, dat%time_Data,
     $      dat%ID_MissVariable, sv%ID_Variable,     
     $      imRNA, ix, bound_data,
     $      network, network_attribute, max_num_edge, num_edge
!      
     $ )
!
!
! --- Shift prior distribution on Reaction Elements ---
!
      Output_File = 'm2constant.log'
      call Output_M2Constant(
!      
     $      mdl%Para_Reg_Modules, mdl%List_Reg_Modules, mdl%strModules,
     $      mdl%Para_Reg_Elements, mdl%List_Reg_Elements, mdl%strElements,
     $      max_num_Reg_Modules, mdl%num_Reg_Modules, max_tmp, 
     $      max_num_Reg_Elements, mdl%num_Reg_Elements,     
     $      imRNA, max_num_Variable, sv%num_Variable,
     $      num_transcript, max_num_transcript, switch, itranscript,
     $      List_Reg_Elements_switch, Output_File
!     
     $  )
!     
!
! --- Set prior distribution to the model parameters ---
!
      iprint = 1      
      call Set_Priors2Model(mdl, pri, sv, iprint)
!
      call get_bound_variable(
!      
     $      mdl%Para_Reg_Modules, mdl%List_Reg_Modules, mdl%strModules,
     $      max_num_Reg_Modules, mdl%num_Reg_Modules, max_tmp,
     $      sv%Name_Variable, sv%StateVariable, 
     $      max_time_StateVariable, max_num_Variable, sv%num_Variable,
     $      itranslation, icomplex, imRNA, 
     $      value_maxminData, RangeHidden,
     $      bound_data, bound_model     
!
     $  )
!
!
      discount_switch = 0.1d0
      call Shift_Prior_Elements_SpecficRegions(
!
     $      pri%Prior_ParaElements, pri%ID_Prior_ParaElements, 
     $      max_num_Para1, pri%num_Para1,
     $      mdl%Para_Reg_Elements, mdl%List_Reg_Elements, 
     $      max_num_Reg_Elements, mdl%num_Reg_Elements,
     $      bound_model, sv%num_Variable, max_num_Variable,
     $      discount_switch
!                  
     $  )     
!
!
! --- Shift noise variances ---
!
      call covariance_matrix(
     $      cov_data, ave_data, resid, 
     $      dat%time_Data, sv%num_Variable, max_time_Data, max_num_Variable
     $  )
!
      call Shift_Noise_Variances(
!      
     $      Variance_Measure, imRNA, dat%ID_MissVariable, 
     $      cov_data, max_num_Variable, sv%num_Variable
!     
     $  )
!
!
! --- Set prior distribution to the model parameters ---
!
      iprint = 1      
      call Set_Priors2Model(mdl, pri, sv, iprint)
!
      if( Nid. EQ. 0 ) then
        call Wrap_Output_Reactions
     $    (mdl, sv, time_interval, pri, Variance_System, Variance_Measure
     $    , trim(mdl%suffix)//'_replaced.emdl')
      endif     

! --- Get uppew/lower bounds of endogeneous variables for the specified model ---
!     
      call get_bound_variable(
!      
     $      mdl%Para_Reg_Modules, mdl%List_Reg_Modules, mdl%strModules,
     $      max_num_Reg_Modules, mdl%num_Reg_Modules, max_tmp,
     $      sv%Name_Variable, sv%StateVariable, 
     $      max_time_StateVariable, max_num_Variable, sv%num_Variable,
     $      itranslation, icomplex, imRNA, 
     $      value_maxminData, RangeHidden,
     $      bound_data, bound_model     
!
     $  )
!
!
! --- Output list of mRNAs and their regulators (activator/repressor) with threshold levels ---
!
      if( Nid. EQ. 0 ) then
!
      Output_File = 'm2constant.log'
      call Output_M2Constant(
!      
     $      mdl%Para_Reg_Modules, mdl%List_Reg_Modules, mdl%strModules,
     $      mdl%Para_Reg_Elements, mdl%List_Reg_Elements, mdl%strElements,
     $      max_num_Reg_Modules, mdl%num_Reg_Modules, max_tmp, 
     $      max_num_Reg_Elements, mdl%num_Reg_Elements,     
     $      imRNA, max_num_Variable, sv%num_Variable,
     $      num_transcript, max_num_transcript, switch, itranscript,     
     $      List_Reg_Elements_switch, Output_File
!     
     $  )
!
      endif
!
!
      if( Nid. EQ. 0 ) then
!
      Output_File = 'bound_model.log'
      call OutputMatrix( bound_model, 2, sv%num_Variable, 
     $          2, max_num_Variable, Output_File )
!
      Output_File = 'bound_data.log'
      call OutputMatrix( bound_data, 2, sv%num_Variable, 
     $          2, max_num_Variable, Output_File )             
!
      endif
!
!
! --- Initialization of matrices and vectors ---
!
      call ArrInit( Cov_ParaModules, max_num_Reg_Modules, 
     $                  max_num_Reg_Modules, 0.0d0 )
      call ArrInit( Cov_ParaElements, max_num_Reg_Elements, 
     $                  max_num_Reg_Elements, 0.0d0 )
      call VecInit( Ave_ParaElements, max_num_Reg_Elements, 0.0d0 )
      call VecInit( Ave_ParaModules, max_num_Reg_Modules, 0.0d0 )
      call VecInit( saveLik, nsave_Particle, 0.0d0 )
      !call ArrInit( Particle_ParaModules, nsave_Particle, max_num_Reg_Modules, 1.0d0 )
      !call ArrInit( Particle_Inits, nsave_Particle, max_num_Variable, 0.0d0 )
      forall (k=1:nsave_Particle)
        elite%o(k)%Particle_ParaModules = 0.0d0
        elite%o(k)%Particle_Inits = 0.0d0
      end forall

! --- Initialization of number of overlaps for the point mass functions in DPM sampling 
!
      do i = 1, nsave_Particle
      num_overlap_DP(i) = 1
      enddo      
!
!
! --- Generate initial simulation trajectory using initial parameters ---
!
!     (Print progress)
!     ------------------------------------------------------------------
      if( Nid. EQ. 0 ) then
      if( iPrint. EQ. 1 ) then      
      write(*,*)
      write(*,'(a)') 
     $  '-@ Initial state variables are specified before 
     $  the parameter search'
      if( iUseDataInits.eq.1 ) then
      write(*,'(a)') 'Set initial observed data at the starting time to
     $  the initial state varaibles [sv%StateVariable(1,*)] and
     $  the prior means [pri%Prior_Inits(*,1)]'
      endif
      endif
      endif
!     ------------------------------------------------------------------      
      n = 5000
      call CPU_TIME( bTime )      
      do i = 1, n
        !/ Run initial simulation
        call Wrap_Run_BioSimulator
     $    (mdl, sv, time_interval, includeNoise, ix, Variance_System)
      end do
      call CPU_TIME( eTime )
      write(*,*) 'version 1', (eTime - bTime)      

      call CPU_TIME( bTime )      
      do i = 1, n
      !/ Run initial simulation
      call Wrap_Run_BioSimulator_version3
     $    (mdl, sv, time_interval, includeNoise, ix, Variance_System)
      end do
      call CPU_TIME( eTime )
      write(*,*) 'version 3', (eTime - bTime)       

!      mdl%nn = 0
!      do i = 1, mdl%num_Reg_Elements
!        ide = mdl%List_Reg_Elements(i,1)
!        do j = 1, mdl%num_Reg_Modules
!          nc = mdl%List_Reg_Modules(j,4)
!          do ii = 1, nc
!            if ( mdl%List_Reg_Modules(j,4+ii).eq.ide ) then
!              mdl%nn = mdl%nn + 1
!              mdl%map_e2m(1,mdl%nn) = i
!              mdl%map_e2m(2,mdl%nn) = j
!            end if
!          end do
!        end do
!      end do
!!
      call Setup_Run_BioSimulator_version2(mdl)
      call CPU_TIME( bTime )
      do i = 1, n
        call Run_BioSimulator_version2
     $    (mdl, sv, time_interval, includeNoise, ix, Variance_System)
      end do
      call CPU_TIME( eTime )
      write(*,*) 'version 2', (eTime - bTime)

      call Setup_Run_BioSimulator_version2a(mdl)
      call CPU_TIME( bTime )
      do i = 1, n
        call Run_BioSimulator_version2a
     $    (mdl, sv, time_interval, includeNoise, ix, Variance_System)
      end do
      call CPU_TIME( eTime )
      write(*,*) 'version 2a', (eTime - bTime)

      !call mpi_finalize(ierr)
      !stop

!     (Print progress)
!     ------------------------------------------------------------------
      if( Nid. EQ. 0 ) then
      if( iPrint. EQ. 1 ) then
      write(*,*)
      write(*,'(a)') 
     $  '-@[Run_BioSimulator_version2] 
     $      Simulation from the initial parameters'
      endif
      endif
!     ------------------------------------------------------------------      
!      / Save the simulation trajectory obtined from the initial parameters at 'simulation_initial.log'
!
      if( Nid. EQ. 0 ) then
        call Wrap_Output_BioSimulator(sv, 'simulation_initial.log')
      endif     

!     (Print progress)
!     ------------------------------------------------------------------
      if( Nid. EQ. 0 ) then
      if( iPrint. EQ. 1 ) then      
      write(*,*)
      write(*,'(a)') 
     $  '-@[Output_BioSimulator] Output result of initial simulation'
      write(*,'(a)') 
     $  '-Access: [simulation_initial.log]'     
      endif
      endif
!     ------------------------------------------------------------------      
!
!
! --- Output list of mRNAs and their regulators (activator/repressor) with threshold levels ---
!
      if( Nid. EQ. 0 ) then
!      
      Output_File = 'm2constant_initial.log'
      call Output_M2Constant(
!      
     $      mdl%Para_Reg_Modules, mdl%List_Reg_Modules, mdl%strModules,
     $      mdl%Para_Reg_Elements, mdl%List_Reg_Elements, mdl%strElements,
     $      max_num_Reg_Modules, mdl%num_Reg_Modules, max_tmp, 
     $      max_num_Reg_Elements, mdl%num_Reg_Elements,     
     $      imRNA, max_num_Variable, sv%num_Variable,
     $      num_transcript, max_num_transcript, switch, itranscript,     
     $      List_Reg_Elements_switch, Output_File
!     
     $  )
!     
      endif
!
      if( iPrint. EQ. 1 ) then      
      write(*,*)
      write(*,'(a)') 
     $  '-@[Output_M2Constant] Output mRNAs and their regulators'
      write(*,'(a)') 
     $  '-Access: [m2constant_initial.log]'
      endif      

!     / Log-likelihood of initial parameters
      valueLik(1) = Wrap_Gauss_Likelihood(dat, sv, Variance_Measure)
      valueLik(1) = dexp( - valueLik(1) / 2.0d0 )
      valueLik(1) = -10d30

!      / Maximum of log-likelihood is defined as that resulting from initial parameters
      valueLik_max = valueLik(1)
!
!     (Print progress)
!     ------------------------------------------------------------------
      if( Nid. EQ. 0 ) then
      if( iPrint. EQ. 1 ) then      
      write(*,*)
      write(*,'(a)')
     $  '-@[Gauss_Likelihood] Likelihood of initial parameters'     
      endif
      endif
!     ------------------------------------------------------------------      
!
!
!     / Save prior (hyper-) parameters as for the base measure of DP 
!$omp parallel
!
!$omp do
      do i = 1, pri%num_Para1
      do j = 1, 4
      priG0%Prior_ParaElements(i,j) = pri%Prior_ParaElements(i,j)
      enddo
      enddo
!$omp end do
!
!$omp do
      do i = 1, pri%num_Para2
      do j = 1, 2
      priG0%Prior_ParaModules(i,j) = pri%Prior_ParaModules(i,j)
      enddo
      enddo
!$omp end do      
!
!$omp do
      do i = 1, sv%num_Variable
      do j = 1, 2
      priG0%Prior_Inits(i,j) = pri%Prior_Inits(i,j)
      enddo
      enddo
!$omp end do
!
!$omp end parallel      
!
!
      call reaction_lists(
!      
     $      mdl%List_Reg_Modules, mdl%num_Reg_Modules, max_num_Reg_Modules,
     $      mdl%strModules, pri%ID_Prior_ParaModules, max_num_Para2, pri%num_Para2,
     $      mdl%List_Reg_Elements, mdl%num_Reg_Elements, max_num_Reg_Elements,
     $      mdl%strElements, pri%ID_Prior_ParaElements, max_num_Para1,pri%num_Para1,
     $      pri%ID_Prior_Inits, sv%num_Variable, max_num_Variable,
     $      list_m, list_e, list_i,
     $      list_mm, list_ee, list_ii
!     
     $ )
!
!$omp parallel
!
!$omp do
      do i = 1, mdl%num_Reg_Elements
      do j = 1, 10
      pelement_map(i,j) = mdl%Para_Reg_Elements(i,j)
      enddo
      enddo
!$omp end do
!
!$omp do
      do i = 1, mdl%num_Reg_Modules
      do j = 1, 10
      pmodule_map(i,j) = mdl%Para_Reg_Modules(i,j)
      enddo
      enddo
!$omp end do
!
!$omp do
      do i = 1, sv%num_Variable
      st_map(i,1) = sv%StateVariable(i,1)
      enddo
!$omp end do
!      
!$omp end parallel         
!
!
! --- Initialization of computational time evaluation ---
!
      if( iDoClock. EQ. 1 ) call system_clock( count_rate = ihz )
!
!
      if( Nid. EQ. 0 ) then
      write( *, * )
      write( *, '(a)' ) '-- @LiSDAS parameter search start --'
      endif
!
!
! (MPI code)
!============================================================
!
!      CALL MPI_Comm_size(MPI_COMM_WORLD, Nproc, ierr)
!     CALL MPI_Comm_rank(MPI_COMM_WORLD, Nid, ierr)
!
!============================================================
!
      ix = Nid + 1000
      imod = mod (num_Particle, Nproc)
      npar_pe = (num_Particle - imod) / Nproc      

! ----- Start Loop (1) for parameter search -----
!      call fpcoll_start()

      do Itr_Particle = 2, npar_pe
!      / Begin computational time evalution
      if( iDoClock. EQ. 1) call system_clock( count = begin_clock )
! --------------------------------------------------------------------------
!$omp   parallel firstprivate(sv, mdl, switch, bound_model, i, j)
!$omp   do
! --------------------------------------------------------------------------
! ----- START Section: Generate particles around the selected base measure
      do k = 1, num_Particle_DPM
!$      id = omp_get_thread_num()
!        / Sampling from the prior on initial state variables
        if( iDoSampleInits. EQ. 1 ) then
          call RandomSampling_Inits (
!            -- Replaced values for intial state variables are stored 
     $      sv%StateVariable, sv%num_Variable,
     $      max_time_StateVariable, max_num_Variable,
!            -- Seed for random number generator
     $      ix,
!            -- Prior distributions
     $      pri%Prior_Inits, pri%ID_Prior_Inits,
!            -- Range of hidden regulatory variables
     $      iDoRangeHidden, RangeHidden,
!            -- Missing indicator
     $      dat%ID_MissVariable
     $    )
        else if (iDoSampleInits. EQ. 2 ) then
          call RandomSampling_Inits_NormalTruncated (
!            -- Replaced values for intial state variables are stored 
     $       sv%StateVariable, sv%num_Variable,
     $       max_time_StateVariable, max_num_Variable,
!            -- Seed for random number generator
     $       ix,
!            -- Prior distributions
     $       pri%Prior_Inits, pri%ID_Prior_Inits,
!            -- Range of hidden regulatory variables
     $       iDoRangeHidden, RangeHidden,
!            -- Range of hidden regulatory variables
     $       bound_data,     
!            -- mRNA indicator
     $       imRNA     
     $    )
        endif

!        / Sampling from the prior on parameters in reaction modules
        if (iDoSampleModule. EQ. 1) then
          call RandomSampling_Modules_S2(
!            -- Replaced values for parameters in Reaction Modules are stored 
     $            mdl%Para_Reg_Modules, mdl%List_Reg_Modules, mdl%num_Reg_Modules, 
     $            max_num_Reg_Modules, max_tmp,
!            -- Seed for random number generator
     $            ix,
!            -- Prior distributions
     $            pri%Prior_ParaModules, pri%ID_Prior_ParaModules, 
     $            pri%num_Para2, max_num_Para2
     $         )
        else if (iDoSampleModule. EQ. 2) then
          scale_variance_module = 0.01d0
          method_sample = 0
          call RandomSampling_Modules_AutoS2Rna(
!            -- Replaced values for parameters in module [mdl%Para_Reg_Modules] and relevant parameters 
     $            mdl%Para_Reg_Modules, mdl%List_Reg_Modules, 
     $            mdl%num_Reg_Modules, max_num_Reg_Modules, max_tmp,
!            -- Seed for random number generator
     $            ix,
!            -- Prior distributions
     $            pri%Prior_ParaModules, pri%ID_Prior_ParaModules,
     $            pri%num_Para2, max_num_Para2,
!            -- Additional parameters
     $            degrad_RNA, synthesis_RNA, 
     $      max_num_Variable, sv%num_Variable,
     $            dat%ID_MissVariable, mdl%strModules, 
!            -- Noise control parameters in generating particles relevant to transcriprion parameters
     $          scale_variance_module, method_sample, 
     $      input_zero_baseline
     $      )
        else if (iDoSampleModule. EQ. 3) then
          call RandomSampling_Modules_AutoMaxRna(
!            -- Replaced values for intial state variables are stored 
     $            mdl%Para_Reg_Modules, mdl%List_Reg_Modules, mdl%num_Reg_Modules, 
     $            max_num_Reg_Modules, max_tmp,
!            -- Seed for random number generator
     $            ix, 
!            -- Prior distributions
     $            pri%Prior_ParaModules, pri%ID_Prior_ParaModules, 
     $            pri%num_Para2, max_num_Para2,
!            -- 
     $            value_maxminData, loc_maxmin, max_num_Variable, sv%num_Variable,
     $            dat%ID_MissVariable, mdl%strModules,
!            -- Noise control parameters in generating particles relevant to transcriprion parameters
     $            scaleUnifNoiseDegRNA, isample_G0, itranslation, icomplex
     $    )
        end if

        if (runif(ix).ge.0.4d0) then
          iDoSampleElement = 1
        else
          iDoSampleElement = 3
        end if

        if (iDoSampleElement. EQ. 1) then
!         / Sampling from prior on parameters in reaction elements
          scale_variance_element = 0.01d0
          call RandomSampling_Elements(
!            -- Replaced values for parameters in elements [mdl%Para_Reg_Elements] and related variables
     $            mdl%Para_Reg_Elements, mdl%List_Reg_Elements, 
     $            mdl%num_Reg_Elements, max_num_Reg_Elements,
!            -- Seed for random number generator
     $            ix, 
!            -- Prior distributions
     $            pri%Prior_ParaElements, pri%ID_Prior_ParaElements, 
     $            pri%num_Para1, max_num_Para1, scale_variance_element
     $      )
        else if (iDoSampleElement. EQ. 2) then
!         / Sampling from prior on parameters in reaction elements
          call RandomSampling_Elements_S2(
!            -- Replaced values for initial state variables are stored 
     $            mdl%Para_Reg_Elements, mdl%List_Reg_Elements, 
     $            mdl%num_Reg_Elements, max_num_Reg_Elements,
!            -- Seed for random number generator
     $            ix, 
!            -- Prior distributions
     $            pri%Prior_ParaElements, pri%ID_Prior_ParaElements, 
     $            pri%num_Para1, max_num_Para1
     $      )
        else if (iDoSampleElement. EQ. 3) then
          call get_bound_variable_inLoop(
     $           mdl%Para_Reg_Modules, mdl%List_Reg_Modules, mdl%strModules,
     $           max_num_Reg_Modules, mdl%num_Reg_Modules, max_tmp,
     $           sv%Name_Variable, max_num_Variable, sv%num_Variable,
     $           itranslation, icomplex, imRNA, 
     $           value_maxminData, RangeHidden,     
     $           bound_data, bound_model, ix
     $         )
          method = 0
          call RandomSampling_Elements_S4(
!            -- Replaced values for intial state variables are stored 
     $            mdl%Para_Reg_Elements, mdl%List_Reg_Elements, mdl%num_Reg_Elements, 
     $            max_num_Reg_Elements, max_tmp,
     $            itranscript, switch, List_Reg_Elements_switch,
     $            num_transcript, max_num_transcript, bound_model,
     $            max_num_Variable,
!            -- Seed for random number generator
     $            ix, 
!            -- Prior distributions
     $            pri%Prior_ParaElements, pri%ID_Prior_ParaElements, 
     $            pri%num_Para1, max_num_Para1, method
     $         )
        else if (iDoSampleElement. EQ. 4 ) then
!      / Run initial simulation
          call Run_BioSimulator_version2
     $      (mdl, sv, time_interval, includeNoise, ix, Variance_System)
          call Get_Switch(
     $           sv%StateVariable, bn_gradient, sv%num_Variable, 
     $           sv%time_StateVariable, max_num_Variable, 
     $           max_time_StateVariable, id,
     $           itranscript, switch, num_transcript, 
     $           max_num_transcript, RefState
     $         )
          call RandomSampling_Elements_S3(
!            -- Replaced values for initial state variables are stored 
     $           mdl%Para_Reg_Elements, mdl%List_Reg_Elements, mdl%num_Reg_Elements, 
     $           max_num_Reg_Elements, max_tmp,
     $           itranscript, switch, List_Reg_Elements_switch,
     $           num_transcript, max_num_transcript,
!            -- Seed for random number generator
     $           ix, 
!            -- Prior distributions
     $           pri%Prior_ParaElements, pri%ID_Prior_ParaElements, 
     $           pri%num_Para1, max_num_Para1
     $         )
        end if
! ----- END Section: Generate particles around the selected base measure

! ----- START Section: Run simulation & evaluate likelihood ----
        call Run_BioSimulator_version2
     $    (mdl, sv, time_interval, includeNoise, ix, Variance_System)

!      / Compute negative log-likelihood with Gaussian distribution
        weight_DPM(k) = Wrap_Gauss_Likelihood(dat, sv, Variance_Measure)
        weight_DPM(k) = dexp(-weight_DPM(k)/2.0d0)

!     / Save particles
        call Save_Model2Particle(mdl, sv, dp%o(k))
      end do 
! --------------------------------------------------------------------------
!$omp end do
!$omp end parallel
! --------------------------------------------------------------------------
!      / Evaluate likelihood of 'pi' by the mean of Monte Calo samples of 'theta'
      call ave(weight_DPM, num_Particle_DPM, wnew)
      call get_max(weight_DPM, num_Particle_DPM, value_max_dpm, loc_max_dpm) 
      valueLik(Itr_Particle) = wnew

      if (Nid .EQ. 0)
     $   write(*,'(i,a1,e10.4)') Itr_Particle, tab, valueLik(Itr_Particle)
! ----- END Section: Run simulation & evaluate likelihood ----


! ----- START Section: update posterior means and covariance matrices -----
      if (iDoSaveCovar. EQ. 1) then
!      / Update posterior means and covariance matrix for parameters in reaction modules
        call SequentialAve(
     $          Ave_ParaModules, sumLik, wnew, mdl%Para_Reg_Modules, 
     $          mdl%num_Reg_Modules, max_num_Reg_Modules, max_tmp 
     $      )
        call SequentialCovar( 
     $          Cov_ParaModules, Ave_ParaModules, sumLik, wnew, 
     $          mdl%Para_Reg_Modules, mdl%num_Reg_Modules, max_num_Reg_Modules, 
     $          max_tmp 
     $      )
!      / Update posterior means and covariance matrix for parameters in reaction elements
        call SequentialAve(
     $          Ave_ParaElements, sumLik, wnew, mdl%Para_Reg_Elements, 
     $          mdl%num_Reg_Elements, max_num_Reg_Elements, max_tmp 
     $      )
        call SequentialCovar( 
     $          Cov_ParaElements, Ave_ParaElements, sumLik, wnew, 
     $          mdl%Para_Reg_Elements, mdl%num_Reg_Elements, max_num_Reg_Elements, 
     $          max_tmp 
     $      )
     
!      / Output posterior means and covariance matrix for parameters in reaction modules
        Output_File = 'save_covari_modules.log'
        call OutputMatrix( 
     $          Cov_ParaModules, mdl%num_Reg_Modules, mdl%num_Reg_Modules, 
     $          max_num_Reg_Modules, max_num_Reg_Modules, Output_File 
     $      )
!      / Output posterior means and covariance matrix for parameters in reaction elements
        Output_File = 'save_covari_elements.log'
        call OutputMatrix( 
     $          Cov_ParaElements, mdl%num_Reg_Elements, mdl%num_Reg_Elements, 
     $          max_num_Reg_Elements, max_num_Reg_Elements, Output_File 
     $      )
      endif
! ----- END Section: update for posterior means and covariance matrices -----


! ----- START Section: Replacement of elite particles having larger weights -----
      if (iDoSaveParticle .EQ. 1 .AND.
     $    valueLik(Itr_Particle) .GT. saveLik(locSaveLik)) then
        saveLik(locSaveLik) = valueLik(Itr_Particle)
        icount_replace_base = icount_replace_base + 1
        num_overlap_DP(locSaveLik) = 1
        num_replace_dpm = num_replace_dpm + 1
        itr_replace_dpm(num_replace_dpm) = Itr_Particle

!       / Output results of elite selection
        if (Nid. EQ. 0) then
          write(*,'(a,a1,a,a1,i6,a1,a,a1,i6,a1,a,a1,e15.4)') 
     $      '@[save particle]', tab, 
     $      '-particle', tab, Itr_Particle, tab, 
     $      '-address', tab, locSaveLik, tab, 
     $      '-likelihood', tab, saveLik(locSaveLik)
        end if

        call Save_Prior2Particle(pri, mdl, sv, elite%o(locSaveLik))
!       / update the worst particle
        call get_min(saveLik, nsave_Particle, value_min, locSaveLik)
      end if
! ----- END Section: Replacement of elite particles having larger weights -----

! ----- START Section: Resampling 1 particle around the selected base measure ----
!     / Normalization of weights (likelihood)
      sum = 0.0d0     
!$omp parallel
!$omp do reduction(+:sum)
      do i = 1, num_Particle_DPM
        sum = sum + weight_DPM(i)
      end do
!$omp end do
!$omp do
      do i = 1, num_Particle_DPM
        pr_DPM(i) = weight_DPM(i)/sum
      end do
!$omp end do
!$omp end parallel

      call Resampling(pr_DPM, num_Particle_DPM, max_num_Particle_DPM, ix, ismp_Particle, 1)
      ir = ismp_Particle(1)
      if (Itr_Particle. GT. numDoDPM) then
        call dGauss_Particles_from_Center(prs, elite, dp%o(ir), mdl, sv, pri)
      end if
! ----- END Section: Resampling 1 particle around the selected base measure ----

!      / Stop time measurement 
      if (iDoClock. EQ. 1) call system_clock( count = end_clock )

!      / Compute CPU time for each unit of the search process
      CompTime(Itr_Particle) = real(end_clock - begin_clock)/real(ihz)
      sumCompTime = sumCompTime + CompTime(Itr_Particle)

! ----- START Section: Update best particle -----
      Nid_map = itest_map_update(value_max_dpm, valueLik_max, Nid)
      if (Nid_map .EQ. Nid) then
        write(*, '(a,a1,a,a1,i8,a1,a,a1,e15.4,a1,a,i4)') 
     $    '@[map]', tab, '-particle', tab, Itr_Particle, tab,
     $    '-likelihood', tab, value_max_dpm, tab, '-Nid', Nid
        num_replace_map = num_replace_map + 1
        itr_replace_map(num_replace_map) = Itr_Particle      
      
!     / Save the current map particle
        Output_File = 'map_'//trim(sI(Itr_Particle))//'_'//trim(sI(Nid))
        call Write_Particle
     $        (dp%o(loc_max_dpm), mdl, sv, time_interval, pri,.true.
     $        ,Output_File, includeNoise, ix, Variance_System, Variance_Measure)

        Output_File = 'm2constant'//'_'//trim(sI(Itr_Particle))//'_'//trim(sI(Nid))//'.log'
        call Output_M2Constant(
     $         mdl%Para_Reg_Modules, mdl%List_Reg_Modules, mdl%strModules,
     $         mdl%Para_Reg_Elements, mdl%List_Reg_Elements, mdl%strElements,
     $         max_num_Reg_Modules, mdl%num_Reg_Modules, max_tmp, 
     $         max_num_Reg_Elements, mdl%num_Reg_Elements,     
     $         imRNA, max_num_Variable, sv%num_Variable,
     $         num_transcript, max_num_transcript, switch, itranscript,     
     $         List_Reg_Elements_switch, Output_File
     $       )
        !collect a snapshot of the computer for speed-test.f90
        call Save_Snapshot("state.bin", dat, mdl, sv, time_interval, Variance_System)
      endif
! ----- END Section: Update best particle -----

! ----- START Section: Dirichlet Process Mixture -----
      if( iDoDPM. EQ. 1 ) then 

	  pr_G0 = gamma/(gamma + dble(numDoDPM))
	  rand = runif(ix)

	  if (Itr_Particle. LE. numDoDPM) pr_G0 = 1.0d0

  !      / Set priors from the initially-specified base measure distribution      
	  if (rand. LE. pr_G0) then
	    isample_G0 = 0

	    if (Itr_Particle. LE. numDoDPM) then
		scale_variace_G0 = 0.1d0
		method = 2
		call Sampling_PriorG0(pri, priG0, sv, ix, method)! 'pri' contains the results.
!
            iprint = 0
            call Set_Priors2Model(mdl, pri, sv, iprint)
!
            call get_bound_variable_inLoop(
     $             mdl%Para_Reg_Modules, mdl%List_Reg_Modules, mdl%strModules,
     $             max_num_Reg_Modules, mdl%num_Reg_Modules, max_tmp,
     $             sv%Name_Variable, max_num_Variable, sv%num_Variable,
     $             itranslation, icomplex, imRNA, 
     $             value_maxminData, RangeHidden,     
     $              bound_data, bound_model, ix
     $           )
!
            method = 1
            call RandomSampling_Elements_S4(
!            -- Replaced values for intial state variables are stored 
     $             mdl%Para_Reg_Elements, mdl%List_Reg_Elements, mdl%num_Reg_Elements, 
     $             max_num_Reg_Elements, max_tmp,
     $             itranscript, switch, List_Reg_Elements_switch,
     $             num_transcript, max_num_transcript, bound_model,
     $             max_num_Variable,   
!            -- Seed for random number generator
     $            ix, 
!            -- Prior distributions
     $            pri%Prior_ParaElements, pri%ID_Prior_ParaElements, 
     $            pri%num_Para1, max_num_Para1, method
     $      )

      endif     

      if (Itr_Particle. GT. numDoDPM) then
      method = 1
      call sampling_priors(pri, priG0, sv, mdl, ix, dp%o(ir), method)

      iprint = 0
      call Set_Priors2Model(mdl, pri, sv, iprint)

      call Run_BioSimulator_version2
     $    (mdl, sv, time_interval, includeNoise, ix, Variance_System)

      call Get_Switch(
     $      sv%StateVariable, bn_gradient, sv%num_Variable, 
     $      sv%time_StateVariable, max_num_Variable, 
     $      max_time_StateVariable, id,
     $      itranscript, switch, num_transcript, 
     $      max_num_transcript, RefState
     $     )          
      endif

      call RandomSampling_Elements_S5(
!            -- Replaced values for intial state variables are stored 
     $            mdl%Para_Reg_Elements, mdl%List_Reg_Elements, mdl%num_Reg_Elements, 
     $            max_num_Reg_Elements, max_tmp,
     $      itranscript, switch, List_Reg_Elements_switch,
     $      num_transcript, max_num_transcript,     
!            -- Seed for random number generator
     $            ix, 
!            -- Prior distributions
     $            pri%Prior_ParaElements, pri%ID_Prior_ParaElements, 
     $            pri%num_Para1, max_num_Para1                                               
     $      )      
      else

      isample_G0 = 0
!      / Normalization of weights
      sum = 0.0d0
!$omp parallel
!$omp do reduction(+:sum)
      do i = 1, nsave_Particle
        weight4resample(i) = prs(i)
        sum = sum + weight4resample(i)
      enddo
!$omp end do
!$omp do
      do i = 1, nsave_Particle
      weight4resample(i) = weight4resample(i) / sum
      enddo
!$omp end do
!
! (OpenMP END)
!------------------------------------------------------------------
!$omp end parallel
!------------------------------------------------------------------
!
!      / Do resampling 
      call Resampling( 
     $          weight4resample, nsave_Particle, 
     $          nsave_Particle, ix, ismp_Particle, 1 
     $  )
!
      ismp = ismp_Particle(1)
      num_overlap_DP(ismp) = num_overlap_DP(ismp) + 1      
!
      if( Nid. EQ. 0 ) then
      write(*,'("@[num_overlap_DP]:", 1000( a1, i6, a1, e12.4 ))')
     $  ( ( tab, num_overlap_DP(i), tab, saveLik(i) ), 
     $  i=1,nsave_Particle )
      endif
!
      if( Nid. EQ. 0 ) then
!      / Output resampling process
      write(*,'(a,a1,a,a1,i6,a1,a,a1,e20.8)') 
     $      '@[replication]', tab, '-sample id', tab, ismp, tab, 
     $      '-weight', tab, weight4resample(ismp)
      endif
!
      method = 2
      call sampling_priors(pri, priG0, sv, mdl, ix, dp%o(ir), method)
      endif

      endif
! ----- END Section: Dirichlet Process Mixture -----

      if (mod(Itr_Particle, 100). EQ. 0) then

      write( *, '(a,a1,a,a1,i8,a1,a,a1,e15.4)' ) 
     $          '@[current.map]', tab, '-particle', tab, Itr_Particle, tab, 
     $      '-likelihood', tab, valueLik_max

      write( str, '( i5 )' ) Nid
      str = adjustl( str )
      Output_File = 'computation_time_'//trim( str )//'.log'
      
!     / Report of computational times
      open(5, file=Output_File )
      write(5,'(a,a1,f20.5)') 
     $ 'Total computational time', tab, sumCompTime
      write(5,'(a)') 'Time for each incremental process of particles'
      do i = 2, Itr_Particle
      write(5,'(f10.5)') CompTime(i) 
      enddo
      close(5)

      Output_File = 'replace_map_'//trim( str )//'.log'
!     / Number of replacement of map
      open(5, file=Output_File )
      write(5,'(a,a1,i10,a1,a,a1,e15.5)') 
     $      'Number of replacement of map', tab, num_replace_map,
     $      tab, 'MAP', tab, valueLik_max
      do i = 1, num_replace_map
      write(5,'(i10)') itr_replace_map(i) 
      enddo
      close(5)

      Output_File = 'replace_dpm_'//trim( str )//'.log'
!     / Number of replacement of DPM base distributions
      open(5, file=Output_File)
      write(5,'(a,a1,i10)') 
     $      'Number of replacement of DPM base distributions', 
     $       tab, num_replace_dpm
      do i = 1, num_replace_dpm
      write(5,'(i10)') itr_replace_dpm(i) 
      enddo
      close(5)      
      
      endif

!     / Save particles (DPM basis distributions)
      if (mod(Itr_Particle, 100). EQ. 0 .OR.
     $     (Itr_Particle .LE. 200 .AND. mod(Itr_Particle, 10) .EQ. 0)) then

!      num_demo = num_demo + 1
!      write( str, '( i5 )' ) num_demo
!      str = adjustl(str)
!      call Save_Particle(dp, sv, mdl, num_overlap_DP, str)

!     / Save simulations for the current DPM basis distributions
!
! (OpenMP START)
! --------------------------------------------------------------------------
!$omp   parallel firstprivate(
!$omp&     sv, mdl, 
!$omp&     str, str2, Output_File, i,j,k, bound_model )
! --------------------------------------------------------------------------

!$omp do
      do k = 1, nsave_Particle

!$    id = omp_get_thread_num()
!
      do i = 1, mdl%num_Reg_Elements
      do j = 1,2
      mdl%Para_Reg_Elements(i,j) = dp%o(k)%Particle_ParaElements(i,j)
      enddo
      enddo
!
      do i = 1, mdl%num_Reg_Modules
      mdl%Para_Reg_Modules(i,1) = dp%o(k)%Particle_ParaModules(i)
      enddo
!
      do i = 1, sv%num_Variable
      sv%StateVariable(i,1) = dp%o(k)%Particle_Inits(i)
      enddo
!
!      / Run initial simulation
!      call Run_BioSimulator_version2
!     $    (mdl, sv, time_interval, includeNoise, ix, Variance_System)
!      write( str, '(i5)' ) k
!      write( str2, '(i5)' ) num_demo
!      str = adjustl( str )
!      str2 = adjustl( str2 )
!!$omp critical
!      !Curiously, directly pushing string expression leads to crash
!      Output_File = 'demo_simulation_'//trim(str)//'_'//trim(str2)//'.log'
!      call Wrap_Output_BioSimulator(sv, Output_File)
!!$omp end critical
!!
      enddo
!$omp end do            
!
! (OpenMP END)
! --------------------------------------------------------------------------
!$omp end parallel
! --------------------------------------------------------------------------
!
      nextend = 5000
!      / Run initial simulation
      call Run_BioSimulator_version2
     $    (mdl, sv, time_interval, includeNoise, ix, Variance_System)

      write(str, '(i5)') k
      write(str2, '(i5)') num_demo
      str = adjustl(str)
      str2 = adjustl(str2)
      Output_File = 'demo_map_'//trim(str)//'_'//trim( str2 )//'.log'      
      call Output_BioSimulator(
     $            st_map, sv%num_Variable, nextend,  
     $            nextend, max_num_Variable,
     $            sv%Name_Variable, Output_File
     $      )
      endif  
!
      enddo

!      call fpcoll_stop()

!     / Report of computational time test
      open(5, file='computation_time.log')
      write(5,'(a,a1,f20.5)') 
     $ 'Total computational time', tab, sumCompTime
      write(5,'(a)') 'Time for each incremental process of particles'
      do i = 2, Itr_Particle
      write(5,'(f10.5)') CompTime(i) 
      enddo
      close(5)

      CALL MPI_Finalize(ierr)
                   
      stop
      end
